# -*- coding: utf-8 -*-
"""Nonce class"""

from __future__ import print_function
import uuid
from typing import Tuple, Dict
from dataclasses import dataclass
from acme2certifier.acme_srv.db_handler import DBstore
from acme2certifier.acme_srv.helper import uts_now, load_config


# Custom Exceptions
class NonceError(Exception):
    """Base exception for nonce operations"""

    # pylint: disable=unnecessary-pass
    pass


class ConfigurationError(NonceError):
    """Raised when configuration is invalid"""

    # pylint: disable=unnecessary-pass
    pass


class NonceRepository:
    """Repository class for Nonce operations."""

    def __init__(self, dbstore) -> None:
        self.dbstore = dbstore

    def check_nonce(self, nonce) -> bool:
        return self.dbstore.nonce_check(nonce)

    def delete_nonce(self, nonce) -> None:
        return self.dbstore.nonce_delete(nonce)

    def add_nonce(self, nonce) -> int:
        return self.dbstore.nonce_add(nonce)

    def delete_nonces(self, nonce_list) -> int:
        return self.dbstore.nonce_delete_bulk(nonce_list)

    def search_expired_nonces(self, timestamp: int) -> Tuple[list, list]:
        return self.dbstore.nonce_search_by_timestamp(timestamp)


@dataclass
class NonceConfiguration:
    """Configuration for Nonce operations"""

    validity: int = 86400  # Default validity in seconds (24 hours)


class Nonce(object):
    """Nonce handler"""

    def __init__(self, debug: bool = False, logger: object = None, repo: object = None):
        self.debug = debug
        self.logger = logger
        self.config = NonceConfiguration()
        self.repo = repo or NonceRepository(DBstore(self.debug, self.logger))

    def __enter__(self):
        """Makes ACMEHandler a Context Manager"""
        self._load_configuration()
        return self

    def __exit__(self, *args):
        """Close the connection at the end of the context"""

    def _load_configuration(self):
        """load configuratoion"""

        config_dic = load_config()

        if config_dic:

            try:
                self.config.validity = int(
                    config_dic.get("Nonce", "validity", fallback=self.config.validity)
                )
            except ValueError as err:
                raise ConfigurationError(
                    f"Invalid validity parameter: {config_dic.get('Nonce', 'validity')}"
                ) from err

    def _validate_and_consume_nonce(self, nonce: str) -> Tuple[int, str, str]:
        """Check if nonce exists and delete it (consume)."""
        self.logger.debug("Nonce._validate_and_consume_nonce(%s)", nonce)
        try:
            nonce_chk_result = self.repo.check_nonce(nonce)
        except Exception as err_:
            self.logger.critical("Database error: failed to check nonce: %s", err_)
            nonce_chk_result = False

        if nonce_chk_result:
            try:
                self.repo.delete_nonce(nonce)
            except Exception as err_:
                self.logger.critical("Database error: failed to delete nonce: %s", err_)
            code = 200
            message = None
            detail = None
        else:
            code = 400
            message = "urn:ietf:params:acme:error:badNonce"
            detail = nonce
        self.logger.debug("Nonce._validate_and_consume_nonce() ended with:%s", code)
        return (code, message, detail)

    def _generate_nonce_value(self) -> str:
        """Generate a new nonce value."""
        self.logger.debug("Nonce._generate_nonce_value()")
        return uuid.uuid4().hex

    def check(self, protected_decoded: Dict[str, str]) -> Tuple[int, str, str]:
        """Check nonce (public method, backward compatible)."""
        self.logger.debug("Nonce.check_nonce()")
        if "nonce" in protected_decoded:
            code, message, detail = self._validate_and_consume_nonce(
                protected_decoded["nonce"]
            )
        else:
            code = 400
            message = "urn:ietf:params:acme:error:badNonce"
            detail = "NONE"
        self.logger.debug("Nonce.check_nonce() ended with:%s", code)
        return (code, message, detail)

    def generate_and_add(self) -> str:
        """Generate new nonce and store it (public method, backward compatible)."""
        self.logger.debug("Nonce.generate_and_add()")
        nonce = self._generate_nonce_value()
        self.logger.debug("got nonce: %s", nonce)
        try:
            self.repo.add_nonce(nonce)
        except Exception as err_:
            self.logger.critical("Database error: failed to add new nonce: %s", err_)
        self.logger.debug("Nonce.generate_and_add() ended with:%s", nonce)
        return nonce

    def expire_nonces(self, timestamp: int = uts_now()) -> Tuple[list, list]:
        """Expire nonces based on expiry date."""
        self.logger.debug("Nonce.expire_nonces()")
        nonce_list = []
        total_deleted = 0

        if self.config.validity <= 0:
            self.logger.warning(
                "Nonce.expire_nonces() skipped: validity is set to %s, which means no expiration.",
                self.config.validity,
            )
            return ([], [])  # Return empty lists as no nonces are expired

        try:
            nonce_list = self.repo.search_expired_nonces(
                timestamp - self.config.validity
            )
            if nonce_list:
                total_deleted = self.repo.delete_nonces(nonce_list)
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to search expired nonces: %s", err_
            )
        self.logger.debug(
            "Nonce.expire_nonces() ended with: %s entries, %s deleted",
            len(nonce_list),
            total_deleted,
        )
        return (
            [],
            nonce_list,
        )  # Return empty list for field_list as it's not used in this context



"""version file"""

# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) pyproject.toml can read it via tool.setuptools.dynamic
# 3) runtime code can import the same value
__version__ = "0.45.dev1"
__dbversion__ = "0.41"



#!/usr/bin/python
"""CLI user management for housekeeping."""

from __future__ import annotations

import argparse
import json
import os.path

from acme2certifier.acme_srv.helper import logger_setup
from acme2certifier.acme_srv.housekeeping import Housekeeping


def arg_parse():
    """simple argparser"""
    parser = argparse.ArgumentParser(
        description="a2c-cliuser-mgmt - manage CLI housekeeping users"
    )
    parser.add_argument(
        "-d", "--debug", help="debug mode", action="store_true", default=False
    )
    parser.add_argument(
        "-c",
        "--certificateadmin",
        help="grant permissions to manage certificates",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--reportadmin",
        help="grant permissions to download reports",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-u",
        "--useradmin",
        help="grant permissions to manage cli users",
        action="store_true",
        default=False,
    )
    parser.add_argument("-e", "--email", help="email address", default=None)
    parser.add_argument("-k", "--keyfile", help="file containing JWK")
    parser.add_argument("-n", "--jwkname", help="name of the key")
    clist = parser.add_mutually_exclusive_group()
    clist.add_argument("--list", help="list users", action="store_true", default=False)
    clist.add_argument(
        "--delete", help="delete user", action="store_true", default=False
    )

    args = parser.parse_args()

    debug = args.debug
    config_dic = {
        "debug": args.debug,
        "permissions": {
            "certificateadmin": args.certificateadmin,
            "reportadmin": args.reportadmin,
            "cliadmin": args.useradmin,
        },
    }

    if args.jwkname:
        config_dic["jwkname"] = args.jwkname
    if args.delete:
        config_dic["delete"] = args.delete
    if args.list:
        config_dic["list"] = args.list
    if args.email:
        config_dic["email"] = args.email
    if args.keyfile:
        try:
            keyfile = validate_keyfile_path(args.keyfile)
            config_dic["jwk"] = json.loads(file_load(keyfile))
        except ValueError as err:
            print(f"Error: {err}")

    return (debug, config_dic)


def validate_keyfile_path(filename):
    """validate keyfile path before reading from disk"""
    if not filename:
        raise ValueError("keyfile path is empty")

    base_dir = os.path.realpath(os.getcwd())
    candidate = os.path.realpath(os.path.abspath(filename))

    if candidate != base_dir and not candidate.startswith(base_dir + os.sep):
        raise ValueError(
            f'Invalid keyfile path "{filename}". Path must be within "{base_dir}".'
        )

    if not os.path.isfile(candidate):
        raise ValueError(f'keyfile "{filename}" does not exist')

    return candidate


def file_load(filename):
    """load file at once"""
    safe_filename = validate_keyfile_path(filename)
    with open(safe_filename, encoding="utf8") as _file:
        lines = _file.read()
    return lines


def main() -> None:
    """Manage CLI housekeeping users."""
    debug, config_dic = arg_parse()
    # the cli program needs to be chatty
    config_dic["silent"] = False
    logger = logger_setup(debug)
    with Housekeeping(debug, logger) as housekeeping:
        housekeeping.cli_usermgr(config_dic)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for nonce.py"""

# pylint: disable=C0302, C0415, R0904, R0913, R0914, R0915, W0212
import unittest
import sys
from unittest.mock import patch, MagicMock

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class FakeDBStore(object):
    """face DBStore class needed for mocking"""

    # pylint: disable=W0107, R0903
    pass


class TestNonce(unittest.TestCase):
    """tests for Nonce"""

    acme = None

    def setUp(self):
        """setup unittest"""
        models_mock = MagicMock()
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.acme_srv.nonce import Nonce

        self.nonce = Nonce(False, self.logger)

    def test_001_generate_nonce_value(self):
        """test Nonce._generate_nonce_value() and check if we get something back"""
        self.assertIsNotNone(self.nonce._generate_nonce_value())

    def test_002_generate_and_add(self):
        """test Nonce._generate_and_add() and check if we get something back"""
        self.assertIsNotNone(self.nonce.generate_and_add())

    def test_003_nonce_check(self):
        """test Nonce.check() with missing nonce"""
        self.assertEqual(
            (400, "urn:ietf:params:acme:error:badNonce", "NONE"),
            self.nonce.check({"foo": "bar"}),
        )

    @patch("acme2certifier.acme_srv.nonce.Nonce._validate_and_consume_nonce")
    def test_004_nonce_check(self, mock_validate_and_consume_nonce):
        """test Nonce.check() calls _validate_and_consume_nonce()"""
        mock_validate_and_consume_nonce.return_value = (200, None, None)
        self.assertEqual((200, None, None), self.nonce.check({"nonce": "aaa"}))

    @patch("acme2certifier.acme_srv.nonce.DBstore")
    def test_005_nonce__validate_and_consume_nonce(self, mock_dbstore_class):
        """test Nonce._validate_and_consume_nonce()"""
        # Setup mock to return True for nonce_check
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.nonce_check.return_value = True
        mock_dbstore_instance.nonce_delete.return_value = None
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Create a new nonce instance with the mocked dbstore
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger)

        self.assertEqual((200, None, None), nonce._validate_and_consume_nonce("aaa"))

    @patch("acme2certifier.acme_srv.nonce.DBstore")
    def test_006_nonce_generate_and_add(self, mock_dbstore_class):
        """test Nonce._add() if dbstore.nonce_add raises an exception"""
        # Setup mock to raise exception
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.nonce_add.side_effect = Exception("exc_nonce_add")
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Create a new nonce instance with the mocked dbstore
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger)

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            nonce.generate_and_add()
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to add new nonce: exc_nonce_add",
            lcm.output,
        )

    @patch("acme2certifier.acme_srv.nonce.DBstore")
    def test_007_nonce__validate_and_consume_nonce(self, mock_dbstore_class):
        """test Nonce._validate_and_consume_nonce() if dbstore.nonce_delete raises an exception"""
        # Setup mock: nonce_check returns True, nonce_delete raises exception
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.nonce_check.return_value = True
        mock_dbstore_instance.nonce_delete.side_effect = Exception("exc_nonce_delete")
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Create a new nonce instance with the mocked dbstore
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger)

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            nonce._validate_and_consume_nonce("nonce")
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to delete nonce: exc_nonce_delete",
            lcm.output,
        )

    @patch("acme2certifier.acme_srv.nonce.DBstore")
    def test_008_nonce__validate_and_consume_nonce(self, mock_dbstore_class):
        """test Nonce._validate_and_consume_nonce() if dbstore.nonce_check raises an exception"""
        # Setup mock to raise exception on nonce_check
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.nonce_check.side_effect = Exception("exc_nonce_check")
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Create a new nonce instance with the mocked dbstore
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger)

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            nonce._validate_and_consume_nonce("nonce")
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to check nonce: exc_nonce_check",
            lcm.output,
        )

    def test_009__enter_(self):
        """test enter"""
        self.nonce.__enter__()

    @patch("acme2certifier.acme_srv.nonce.load_config", return_value=None)
    def test_010_load_configuration_without_config(self, _mock_load_config):
        """test _load_configuration() keeps default validity if config is missing"""
        self.nonce.config.validity = 7200
        self.nonce._load_configuration()
        self.assertEqual(7200, self.nonce.config.validity)

    @patch("acme2certifier.acme_srv.nonce.load_config")
    def test_011_load_configuration_with_validity(self, mock_load_config):
        """test _load_configuration() applies Nonce.validity from config"""
        config_mock = MagicMock()
        config_mock.get.return_value = "3600"
        mock_load_config.return_value = config_mock

        self.nonce.config.validity = 7200
        self.nonce._load_configuration()

        self.assertEqual(3600, self.nonce.config.validity)
        config_mock.get.assert_called_once_with("Nonce", "validity", fallback=7200)

    @patch("acme2certifier.acme_srv.nonce.load_config")
    def test_012_load_configuration_invalid_validity(self, mock_load_config):
        """test _load_configuration() raises ConfigurationError on invalid validity"""
        config_mock = MagicMock()
        config_mock.get.return_value = "invalid"
        mock_load_config.return_value = config_mock

        from acme2certifier.acme_srv.nonce import ConfigurationError

        with self.assertRaises(ConfigurationError) as ctx:
            self.nonce._load_configuration()

        self.assertIn("Invalid validity parameter", str(ctx.exception))

    @patch("acme2certifier.acme_srv.nonce.load_config")
    def test_013_load_configuration_uses_current_fallback(self, mock_load_config):
        """test _load_configuration() uses current validity as fallback value"""
        config_mock = MagicMock()
        config_mock.get.return_value = "1800"
        mock_load_config.return_value = config_mock

        self.nonce.config.validity = 999
        self.nonce._load_configuration()

        config_mock.get.assert_called_once_with("Nonce", "validity", fallback=999)
        self.assertEqual(1800, self.nonce.config.validity)

    def test_014_expire_nonces_skips_if_validity_disabled(self):
        """test expire_nonces() returns empty result when validity <= 0"""
        repo_mock = MagicMock()
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger, repo=repo_mock)
        nonce.config.validity = 0

        with self.assertLogs("test_a2c", level="WARNING") as lcm:
            result = nonce.expire_nonces(timestamp=10000)

        self.assertEqual(([], []), result)
        self.assertIn(
            "Nonce.expire_nonces() skipped: validity is set to 0", lcm.output[0]
        )
        repo_mock.search_expired_nonces.assert_not_called()
        repo_mock.delete_nonces.assert_not_called()

    def test_015_expire_nonces_no_expired_entries(self):
        """test expire_nonces() with no expired nonce entries"""
        repo_mock = MagicMock()
        repo_mock.search_expired_nonces.return_value = []
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger, repo=repo_mock)
        nonce.config.validity = 120

        result = nonce.expire_nonces(timestamp=5000)

        self.assertEqual(([], []), result)
        repo_mock.search_expired_nonces.assert_called_once_with(4880)
        repo_mock.delete_nonces.assert_not_called()

    def test_016_expire_nonces_with_expired_entries(self):
        """test expire_nonces() deletes found expired nonces"""
        repo_mock = MagicMock()
        repo_mock.search_expired_nonces.return_value = ["n1", "n2", "n3"]
        repo_mock.delete_nonces.return_value = 3
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger, repo=repo_mock)
        nonce.config.validity = 300

        result = nonce.expire_nonces(timestamp=2000)

        self.assertEqual(([], ["n1", "n2", "n3"]), result)
        repo_mock.search_expired_nonces.assert_called_once_with(1700)
        repo_mock.delete_nonces.assert_called_once_with(["n1", "n2", "n3"])

    def test_017_expire_nonces_handles_search_exception(self):
        """test expire_nonces() handles repository search errors"""
        repo_mock = MagicMock()
        repo_mock.search_expired_nonces.side_effect = Exception("exc_search")
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger, repo=repo_mock)
        nonce.config.validity = 60

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            result = nonce.expire_nonces(timestamp=1000)

        self.assertEqual(([], []), result)
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to search expired nonces: exc_search",
            lcm.output,
        )
        repo_mock.delete_nonces.assert_not_called()

    def test_018_expire_nonces_handles_delete_exception(self):
        """test expire_nonces() handles repository delete errors"""
        repo_mock = MagicMock()
        repo_mock.search_expired_nonces.return_value = ["n1", "n2"]
        repo_mock.delete_nonces.side_effect = Exception("exc_delete")
        from acme2certifier.acme_srv.nonce import Nonce

        nonce = Nonce(False, self.logger, repo=repo_mock)
        nonce.config.validity = 100

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            result = nonce.expire_nonces(timestamp=1000)

        self.assertEqual(([], ["n1", "n2"]), result)
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to search expired nonces: exc_delete",
            lcm.output,
        )
        repo_mock.search_expired_nonces.assert_called_once_with(900)
        repo_mock.delete_nonces.assert_called_once_with(["n1", "n2"])


class TestNonceRepository(unittest.TestCase):
    """tests for NonceRepository pass-through methods"""

    def setUp(self):
        """setup repository test fixtures"""
        from acme2certifier.acme_srv.nonce import NonceRepository

        self.dbstore_mock = MagicMock()
        self.repo = NonceRepository(self.dbstore_mock)

    def test_019_repository_delete_nonces(self):
        """test NonceRepository.delete_nonces() forwards call to DB layer"""
        self.dbstore_mock.nonce_delete_bulk.return_value = 2

        result = self.repo.delete_nonces(["n1", "n2"])

        self.assertEqual(2, result)
        self.dbstore_mock.nonce_delete_bulk.assert_called_once_with(["n1", "n2"])

    def test_020_repository_check_nonce(self):
        """test NonceRepository.check_nonce() forwards call to DB layer"""
        self.dbstore_mock.nonce_check.return_value = True

        result = self.repo.check_nonce("nonce-1")

        self.assertTrue(result)
        self.dbstore_mock.nonce_check.assert_called_once_with("nonce-1")

    def test_021_repository_delete_nonce(self):
        """test NonceRepository.delete_nonce() forwards call to DB layer"""
        self.dbstore_mock.nonce_delete.return_value = None

        result = self.repo.delete_nonce("nonce-2")

        self.assertIsNone(result)
        self.dbstore_mock.nonce_delete.assert_called_once_with("nonce-2")

    def test_022_repository_add_nonce(self):
        """test NonceRepository.add_nonce() forwards call to DB layer"""
        self.dbstore_mock.nonce_add.return_value = 42

        result = self.repo.add_nonce("nonce-3")

        self.assertEqual(42, result)
        self.dbstore_mock.nonce_add.assert_called_once_with("nonce-3")

    def test_023_repository_search_expired_nonces(self):
        """test NonceRepository.search_expired_nonces() forwards timestamp filter"""
        self.dbstore_mock.nonce_search_by_timestamp.return_value = ["n1", "n2"]

        result = self.repo.search_expired_nonces(1234)

        self.assertEqual(["n1", "n2"], result)
        self.dbstore_mock.nonce_search_by_timestamp.assert_called_once_with(1234)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Insta Active Security API  handler"""

from __future__ import print_function
from typing import Tuple, Dict
import os
import requests
from requests.auth import HTTPBasicAuth

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    load_config,
    encode_url,
    csr_pubkey_get,
    csr_cn_get,
    csr_san_get,
    uts_now,
    uts_to_date_utc,
    b64_decode,
    cert_der2pem,
    convert_byte_to_string,
    cert_ski_get,
    config_eab_profile_load,
    config_headerinfo_load,
    eab_profile_header_info_check,
    eab_profile_revocation_check,
    config_enroll_config_log_load,
    config_profile_load,
    enrollment_config_log,
    handler_config_check,
)


class CAhandler(object):
    """EST CA  handler"""

    def __init__(self, _debug: bool = None, logger: object = None):
        self.logger = logger
        self.api_host = None
        self.api_user = None
        self.api_password = None
        self.api_key = None
        self.ca_bundle = None
        self.proxy = None
        self.request_timeout = 10
        self.ca_name = None
        self.auth = None
        self.profile_name = None
        self.cert_validity_days = 30
        self.header_info_field = False
        self.eab_handler = None
        self.eab_profiling = False
        self.enrollment_config_log = False
        self.enrollment_config_log_skip_list = []
        self.profiles = {}
        self.profile_mapping_field = "profile_name"

    def __enter__(self):
        """Makes CAhandler a Context Manager"""
        if not self.api_host:
            self._config_load()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _api_get(self, url: str) -> Tuple[int, Dict[str, str]]:
        """post data to API"""
        self.logger.debug("CAhandler._api_get()")
        headers = {"x-api-key": self.api_key}

        try:
            api_response = requests.get(
                url=url,
                headers=headers,
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            )
            code = api_response.status_code
            try:
                content = api_response.json()
            except Exception as err_:
                self.logger.error(
                    "Could not parse the response for an API get() request: %s", err_
                )
                content = str(err_)
        except Exception as err_:
            self.logger.error("API get() request returned error: %s", err_)
            code = 500
            content = str(err_)

        return code, content

    def _api_post(self, url: str, data: Dict[str, str]) -> Tuple[int, Dict[str, str]]:
        """post data to API"""
        self.logger.debug("CAhandler._api_post()")
        headers = {"x-api-key": self.api_key}

        try:
            api_response = requests.post(
                url=url,
                headers=headers,
                json=data,
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            )
            code = api_response.status_code
            if api_response.text:
                try:
                    content = api_response.json()
                except Exception as err_:
                    self.logger.error(
                        "Could not parse the response for an API post() request: %s",
                        err_,
                    )
                    content = str(err_)
            else:
                content = None
        except Exception as err_:
            self.logger.error("API post() request returned an error: %s", err_)
            code = 500
            content = str(err_)

        return code, content

    def _auth_set(self):
        """set basic authentication header"""
        self.logger.debug("CAhandler._auth_set()")
        if self.api_user and self.api_password:
            self.auth = HTTPBasicAuth(self.api_user, self.api_password)
        else:
            self.logger.error(
                'Auth information incomplete. Either "api_user" or "api_password" parameter is missing in config file'
            )
        self.logger.debug("CAhandler._auth_set() ended")

    def _config_host_load(self, config_dic: Dict[str, str]):
        """load hostname"""
        self.logger.debug("_config_host_load()")

        api_host_variable = config_dic.get("api_host_variable")
        if api_host_variable:
            self.api_host = os.environ.get(api_host_variable)
            if not self.api_host:
                self.logger.error(f"Could not load host_variable: {api_host_variable}")

        api_host = config_dic.get("api_host")
        if api_host:
            if self.api_host:
                self.logger.info("Overwrite api_host parameter")
            self.api_host = api_host

        self.logger.debug("_config_host_load() ended")

    def _certificates_list(self) -> Dict[str, str]:
        """list profiles"""
        self.logger.debug("CAhandler._certificates_list()")

        url = f"{self.api_host}/list_certificates?issuerName={encode_url(self.logger, self.ca_name)}"
        _code, api_response = self._api_get(url)

        self.logger.debug("CAhandler._certificates_list() ended")
        return api_response

    def _config_key_load(self, config_dic: Dict[str, str]):
        """load keyname"""
        self.logger.debug("_config_key_load()")

        api_key_variable = config_dic.get("api_key_variable")
        if api_key_variable:
            self.api_key = os.environ.get(api_key_variable)
            if not self.api_key:
                self.logger.error(f"Could not load key_variable: {api_key_variable}")

        api_key = config_dic.get("api_key")
        if api_key:
            if self.api_key:
                self.logger.info("Overwrite api_key parameter")
            self.api_key = api_key

        self.logger.debug("_config_key_load() ended")

    def _config_password_load(self, config_dic: Dict[str, str]):
        """load passwordname"""
        self.logger.debug("_config_password_load()")

        api_password_variable = config_dic.get("api_password_variable")
        if api_password_variable:
            self.api_password = os.environ.get(api_password_variable)
            if not self.api_password:
                self.logger.error(
                    f"Could not load password_variable: {api_password_variable}"
                )

        api_password = config_dic.get("api_password")
        if api_password:
            if self.api_password:
                self.logger.info("Overwrite api_password parameter")
            self.api_password = api_password

        self.logger.debug("_config_password_load() ended")

    def _config_user_load(self, config_dic: Dict[str, str]):
        """load username"""
        self.logger.debug("_config_user_load()")

        api_user_variable = config_dic.get("api_user_variable")
        if api_user_variable:
            self.api_user = os.environ.get(api_user_variable)
            if not self.api_user:
                self.logger.error(f"Could not load user_variable: {api_user_variable}")

        api_user = config_dic.get("api_user")
        if api_user:
            if self.api_user:
                self.logger.info("Overwrite api_user parameter")
            self.api_user = api_user

        self.logger.debug("_config_user_load() ended")

    def _config_load(self):
        """ " load config from file"""
        self.logger.debug("CAhandler._config_load()")

        config_dic = load_config(self.logger, "CAhandler")

        if "CAhandler" in config_dic:
            self._config_host_load(config_dic["CAhandler"])
            self._config_user_load(config_dic["CAhandler"])
            self._config_password_load(config_dic["CAhandler"])
            self._config_key_load(config_dic["CAhandler"])
            self.ca_name = config_dic["CAhandler"].get("ca_name")
            self.profile_name = config_dic["CAhandler"].get(self.profile_mapping_field)

            if (
                "ca_bundle" in config_dic["CAhandler"]
                and config_dic["CAhandler"]["ca_bundle"] == "False"
            ):
                self.ca_bundle = False
            else:
                self.ca_bundle = config_dic["CAhandler"].get("ca_bundle")

            try:
                self.request_timeout = int(
                    config_dic["CAhandler"].get("request_timeout", 10)
                )
            except Exception as err:
                self.logger.error(
                    "request_timeout parameter is not an integer. Error: %s", err
                )

            try:
                self.cert_validity_days = int(
                    config_dic["CAhandler"].get("cert_validity_days", 30)
                )
            except Exception as err:
                self.logger.error(
                    "cert_validity_days parameter is not an integer. Error: %s", err
                )

        for ele in [
            "api_host",
            "api_user",
            "api_password",
            "api_key",
            "ca_name",
            self.profile_mapping_field,
        ]:
            if not getattr(self, ele):
                self.logger.error(
                    "Configuration incomplete. Variable %s has not been not set", ele
                )

        # load profiling
        self.eab_profiling, self.eab_handler = config_eab_profile_load(
            self.logger, config_dic
        )

        self._auth_set()
        # load enrollment config log
        (
            self.enrollment_config_log,
            self.enrollment_config_log_skip_list,
        ) = config_enroll_config_log_load(self.logger, config_dic)
        # load profiles
        self.profiles = config_profile_load(self.logger, config_dic)
        # load header info
        self.header_info_field = config_headerinfo_load(self.logger, config_dic)

        self.logger.debug("CAhandler._config_load() ended")

    def _csr_cn_get(self, csr: str) -> str:
        """get CN from csr"""
        self.logger.debug("CAhandler._csr_cn_get()")

        cn = csr_cn_get(self.logger, csr)

        if not cn:
            self.logger.info("CN not found in CSR")
            san_list = csr_san_get(self.logger, csr)
            if san_list:
                _type, san_value = san_list[0].split(":")
                cn = san_value
                self.logger.info(
                    "CN not found in CSR. Using first SAN entry as CN: %s",
                    san_value,
                )
            else:
                self.logger.error("CN not found in CSR. No SAN entries found")

        self.logger.debug("CAhandler._csr_cn_get() ended with: %s", cn)
        return cn

    def _issuer_verify(self) -> str:
        """verify issuer"""
        self.logger.debug("CAhandler._issuer_verify()")

        api_response = self._issuers_list()

        if "issuers" in api_response:
            if self.ca_name in api_response["issuers"]:
                error = None
            else:
                error = f"CA {self.ca_name} not found"
                self.logger.error("CAhandler.enroll(): CA %s not found", self.ca_name)
        else:
            error = "Malformed response"
            self.logger.error('Malformed response. "issuers" key not found')

        self.logger.debug("CAhandler._issuer_verify() ended with: %s", error)
        return error

    def _issuers_list(self) -> Dict[str, str]:
        """list issuers"""
        self.logger.debug("CAhandler._list_issuers()")

        url = f"{self.api_host}/list_issuers"
        _code, api_response = self._api_get(url)

        self.logger.debug("CAhandler._list_issuers() ended")
        return api_response

    def _profiles_list(self) -> Dict[str, str]:
        """list profiles"""
        self.logger.debug("CAhandler._profiles_list()")

        url = f"{self.api_host}/list_profiles?issuerName={encode_url(self.logger, self.ca_name)}"
        _code, api_response = self._api_get(url)

        self.logger.debug("CAhandler._profiles_list() ended")
        return api_response

    def _profile_verify(self) -> str:
        """verify profile"""
        self.logger.debug("CAhandler._profile_verify(%s)", self.profile_name)
        api_response = self._profiles_list()

        if "profiles" in api_response:
            if self.profile_name in api_response["profiles"]:
                error = None
            else:
                error = f"Profile {self.profile_name} not found"
                self.logger.error("Profile %s not found", self.profile_name)
        else:
            error = "Malformed response"
            self.logger.error('Malformed response. "profiles" key not found')

        self.logger.debug("CAhandler._profile_verify() ended with: %s", error)
        return error

    def _validity_dates_get(self) -> Tuple[str, str]:
        """calculate validity dates"""
        self.logger.debug("CAhandler._validity_dates_get()")

        uts_now_ = uts_now()
        validfrom = uts_to_date_utc(uts_now_, tformat="%Y-%m-%dT%H:%M:%S")
        validto = uts_to_date_utc(
            uts_now_ + (self.cert_validity_days * 24 * 60 * 60),
            tformat="%Y-%m-%dT%H:%M:%S",
        )

        self.logger.debug("CAhandler._validity_dates_get() ended")
        return validfrom, validto

    def _pem_cert_chain_generate(self, certs_list: list) -> str:
        """generate PEM certificate chain"""
        self.logger.debug("CAhandler._pem_cert_chain_generate()")

        pem_chain = ""
        for cert in certs_list:
            pem_chain += convert_byte_to_string(
                cert_der2pem(b64_decode(self.logger, cert))
            )

        self.logger.debug("CAhandler._pem_cert_chain_generate() ended")
        return pem_chain

    def _issuer_chain_get(self) -> str:
        """get issuer chain"""
        self.logger.debug("CAhandler._issuer_chain_get()")

        url = f"{self.api_host}/get_issuer_chain?issuerName={encode_url(self.logger, self.ca_name)}"
        _code, api_response = self._api_get(url)
        if "certs" in api_response:
            pem_chain = self._pem_cert_chain_generate(api_response["certs"])
        else:
            self.logger.error('"certs" key in issuer chain not found')
            pem_chain = None

        self.logger.debug("CAhandler._issuer_chain_get() ended")
        return pem_chain

    def _cert_get(self, data_dic: Dict[str, str]) -> str:
        """get certificate"""
        self.logger.debug("CAhandler._cert_get()")

        url = f"{self.api_host}/issue_certificate"
        code, api_response = self._api_post(url, data_dic)

        if code == 200 and api_response:
            cert = api_response
        else:
            self.logger.error("Enrollment failed: %s/%s", code, api_response)
            cert = None

        self.logger.debug("CAhandler._cert_get() ended")
        return cert

    def _cert_status_get(self, certificate: str) -> str:
        """get certificate status"""
        self.logger.debug("CAhandler._cert_status_get()")

        data_dic = {"certificateFile": certificate}
        url = f"{self.api_host}/verify_certificate?issuerName={encode_url(self.logger, self.ca_name)}"
        code, api_response = self._api_post(url, data_dic)
        api_response["code"] = code

        return api_response

    def _enrollment_dic_create(self, csr: str) -> Dict[str, str]:
        """create enrollment dic"""
        self.logger.debug("CAhandler._enrollment_dic_create()")

        # get public key from csr
        csr_pubkey = csr_pubkey_get(self.logger, csr, encoding="base64der")
        if csr_pubkey:
            # get CN from csr
            csr_cn = self._csr_cn_get(csr)

            # calculate validiaty dates
            validfrom, validto = self._validity_dates_get()

            # prepare payload for api call
            data_dic = {
                "publicKey": csr_pubkey,
                "profileName": self.profile_name,
                "issuerName": self.ca_name,
                "cn": csr_cn,
                "notBefore": validfrom,
                "notAfter": validto,
            }

            # get SANs from csr as base64 encoded byte sequence
            # sans_base64 = csr_san_byte_get(self.logger, csr)
            # if sans_base64:
            #    data_dic['extensions'] = [{'oid': '2.5.29.17', 'value': sans_base64}]  # 'Zm9vLmJhci5sb2NhbA=='

        else:
            self.logger.error("Could not extract the public key from CSR")
            data_dic = None

        return data_dic

    def enroll(self, csr: str) -> Tuple[str, str, str, str]:
        """enroll certificate"""
        self.logger.debug("CAhandler.enroll()")

        cert_bundle = None
        error = None
        cert_raw = None
        poll_indentifier = None

        # check for eab profiling and header_info
        error = eab_profile_header_info_check(
            self.logger, self, csr, self.profile_mapping_field
        )

        if self.enrollment_config_log:
            self.enrollment_config_log_skip_list.extend(["api_password", "auth"])
            enrollment_config_log(
                self.logger, self, self.enrollment_config_log_skip_list
            )

        if not error:
            # verify issuer
            error = self._issuer_verify()

        if not error:
            # verify profile
            error = self._profile_verify()
            if not error:

                # get issuer chain
                issuer_chain = self._issuer_chain_get()

                data_dic = self._enrollment_dic_create(csr)
                if data_dic:
                    cert_raw = self._cert_get(data_dic)

                if cert_raw:
                    cert = convert_byte_to_string(
                        cert_der2pem(b64_decode(self.logger, cert_raw))
                    )
                    cert_bundle = cert + issuer_chain
                else:
                    error = "Enrollment failed"

        self.logger.debug("Certificate.enroll() ended with: %s", error)
        return (error, cert_bundle, cert_raw, poll_indentifier)

    def handler_check(self):
        """check if handler is ready"""
        self.logger.debug("CAhandler.check()")

        error = handler_config_check(
            self.logger,
            self,
            [
                "api_host",
                "api_user",
                "api_password",
                "api_key",
                "ca_name",
                self.profile_mapping_field,
            ],
        )

        self.logger.debug("CAhandler.check() ended with %s", error)
        return error

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, bool]:
        """poll status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None
        rejected = False

        self.logger.debug("CAhandler.poll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self,
        cert: str,
        _rev_reason: str = "unspecified",
        _rev_date: str = uts_to_date_utc(uts_now()),
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.revoke()")

        code = None
        message = None
        detail = None

        # modify handler configuration in case of eab profiling
        if self.eab_profiling:
            eab_profile_revocation_check(self.logger, self, cert)

        cert_ski = cert_ski_get(
            self.logger, cert
        )  # get subjectKeyIdentifier from certificate
        url = f"{self.api_host}/revoke_certificate?issuerName={encode_url(self.logger, self.ca_name)}&certificateId={cert_ski}"
        data_dic = {}
        code, content_dic = self._api_post(url, data_dic)
        if content_dic:
            message = "urn:ietf:params:acme:error:serverInternal"
            if "Message" in content_dic:
                detail = content_dic.get("Message")
            elif "message" in content_dic:
                detail = content_dic.get("message")
            else:
                detail = "Unknown error"

        self.logger.debug(
            "Certificate.revoke() ended with code: %s, message: %s, detail: %s",
            code,
            message,
            detail,
        )
        return (code, message, detail)

    def trigger(self, _payload: str) -> Tuple[str, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



#!/usr/bin/python
"""database updater"""

from __future__ import annotations

from acme2certifier.acme_srv.db_handler import DBstore
from acme2certifier.acme_srv.helper import logger_setup


def main() -> None:
    """Run SQLite/WSGI database schema update."""
    debug = True
    logger = logger_setup(debug)
    dbstore = DBstore(debug, logger)
    dbstore.db_update()


if __name__ == "__main__":
    main()



"""error.py"""

# pylint: disable=C0209, R1705
from impacket import hresult_errors


def translate_error_code(error_code: int) -> str:
    """translate error code in something readable"""
    error_code &= 0xFFFFFFFF
    if error_code in hresult_errors.ERROR_MESSAGES:
        error_msg_short = hresult_errors.ERROR_MESSAGES[error_code][0]
        error_msg_verbose = hresult_errors.ERROR_MESSAGES[error_code][1]
        return "code: 0x%x - %s - %s" % (
            error_code,
            error_msg_short,
            error_msg_verbose,
        )
    else:
        return "unknown error code: 0x%x" % error_code



import unittest
from unittest.mock import MagicMock, patch, ANY
import os
import sys

# Add the parent directory to sys.path so we can import acme_srv
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

mock_db_handler = MagicMock()
mock_dbstore_class = MagicMock()
mock_db_handler.DBstore = mock_dbstore_class
sys.modules["acme2certifier.acme_srv.db_handler"] = mock_db_handler
from acme2certifier.acme_srv.directory import Directory, DirectoryConfig, DirectoryRepository


class TestDirectory(unittest.TestCase):
    def setUp(self):
        """Set up module-level mocks before any tests run"""
        self.mock_logger = MagicMock()
        self.mock_dbstore = MagicMock()
        self.mock_repository = DirectoryRepository(self.mock_dbstore, self.mock_logger)
        self.mock_config = DirectoryConfig()
        self.mock_cahandler = MagicMock()
        self.mock_cahandler_instance = MagicMock()
        self.mock_cahandler.return_value = self.mock_cahandler_instance
        self.mock_cahandler_instance.__enter__.return_value = (
            self.mock_cahandler_instance
        )
        self.mock_cahandler_instance.__exit__.return_value = None
        self.mock_cahandler_instance.handler_check.return_value = None
        self.directory = Directory(
            debug=None, srv_name="http://localhost", logger=self.mock_logger
        )
        self.directory.dbstore = self.mock_dbstore
        self.directory.repository = self.mock_repository
        self.directory.config = self.mock_config
        self.directory.cahandler = self.mock_cahandler

    def test_001_context_manager(self):
        with patch.object(self.directory, "_load_configuration") as mock_load_config:
            with self.directory as d:
                mock_load_config.assert_called_once()
                self.assertIs(d, self.directory)

    def test_002_load_configuration(self):
        # Mock config_dic to behave like configparser.ConfigParser
        config_dic = MagicMock()
        config_dic.getboolean.return_value = False
        with patch("acme2certifier.acme_srv.directory.load_config", return_value=config_dic):
            with patch(
                "acme2certifier.acme_srv.directory.config_async_mode_load", return_value=False
            ) as mock_async_mode_load:
                with (
                    patch.object(
                        self.directory, "_parse_directory_section"
                    ) as mock_parse_dir,
                    patch.object(self.directory, "_parse_booleans") as mock_parse_bool,
                    patch.object(
                        self.directory, "_parse_eab_and_profiles"
                    ) as mock_parse_eab,
                    patch.object(self.directory, "_load_ca_handler") as mock_load_ca,
                    patch.object(
                        self.directory, "_parse_cahandler_section"
                    ) as mock_parse_cahandler_section,
                ):
                    self.directory._load_configuration()
                    mock_parse_dir.assert_called()
                    mock_parse_bool.assert_called()
                    mock_parse_eab.assert_called()
                    mock_load_ca.assert_called()
                    mock_parse_cahandler_section.assert_called()
                    mock_async_mode_load.assert_called()

    def test_003_parse_directory_empty(self):
        # Mock config_dic to behave like configparser.ConfigParser
        mock_config = MagicMock()
        mock_config.get.side_effect = lambda section, key, fallback=None: {
            ("CAhandler", "foo"): "bar"
        }.get((section, key), fallback)
        self.directory._parse_directory_section(mock_config)
        self.assertFalse(self.directory.config.tos_url)
        self.assertEqual(self.directory.config.url_prefix, "")
        self.assertEqual(
            self.directory.config.home, "https://github.com/grindsa/acme2certifier"
        )

    def test_004_parse_directory_section_sets_config(self):
        # Mock config_dic to behave like configparser.ConfigParser
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "Directory"
        config_dic.__getitem__.side_effect = lambda k: (
            {"tos_url": "tos", "url_prefix": "/prefix", "home": "custom_home"}
            if k == "Directory"
            else None
        )
        config_dic.get.side_effect = lambda section, key, fallback=None: None
        self.directory._parse_directory_section(config_dic)
        self.assertEqual(self.directory.config.tos_url, "tos")
        self.assertEqual(self.directory.config.url_prefix, "/prefix")
        self.assertEqual(self.directory.config.home, "custom_home")

    def test_005_parse_caaidentities_json(self):
        value = '["id1", "id2"]'
        result = self.directory._parse_caaidentities(value)
        self.assertEqual(result, ["id1", "id2"])

    def test_006_parse_caaidentities_fallback(self):
        value = "id1"
        result = self.directory._parse_caaidentities(value)
        self.assertEqual(result, ["id1"])

    def test_007_parse_caaidentities_error(self):
        value = "[invalid_json]"
        with patch.object(self.mock_logger, "error") as mock_error:
            result = self.directory._parse_caaidentities(value)
            self.assertEqual(result, [])
            mock_error.assert_called()

    def test_008_parse_booleans(self):
        config_dic = MagicMock()
        config_dic.getboolean.side_effect = lambda section, key, fallback: True
        self.directory._parse_booleans(config_dic)
        self.assertTrue(self.directory.config.supress_version)
        self.assertTrue(self.directory.config.db_check)
        self.assertTrue(self.directory.config.suppress_product_information)

    def test_009_parse_booleans_error(self):
        config_dic = MagicMock()
        config_dic.getboolean.side_effect = Exception("fail")
        with patch.object(self.mock_logger, "error") as mock_error:
            self.directory._parse_booleans(config_dic)
            self.assertTrue(mock_error.called)

    def test_010_parse_eab_and_profiles(self):
        config_dic = {"EABhandler": {"eab_handler_file": "file"}}
        with patch(
            "acme2certifier.acme_srv.directory.config_profile_load", return_value={"profile": "data"}
        ):
            self.directory._parse_eab_and_profiles(config_dic)
            self.assertTrue(self.directory.config.eab)
            self.assertTrue(self.directory.config.eab_strict_mode)
            self.assertEqual(self.directory.config.profiles, {"profile": "data"})

    def test_011_parse_eab_and_profiles_strict_mode_false(self):
        config_dic = {
            "EABhandler": {"eab_handler_file": "file", "eab_strict_mode": "false"}
        }
        with patch(
            "acme2certifier.acme_srv.directory.config_profile_load", return_value={"profile": "data"}
        ):
            self.directory._parse_eab_and_profiles(config_dic)
            self.assertTrue(self.directory.config.eab)
            self.assertFalse(self.directory.config.eab_strict_mode)

    def test_012_parse_eab_and_profiles_strict_mode_invalid_fallback(self):
        config_dic = {
            "EABhandler": {
                "eab_handler_file": "file",
                "eab_strict_mode": "not-a-bool",
            }
        }
        with patch(
            "acme2certifier.acme_srv.directory.config_profile_load", return_value={"profile": "data"}
        ):
            with patch.object(self.mock_logger, "warning") as mock_warning:
                self.directory._parse_eab_and_profiles(config_dic)
                self.assertTrue(self.directory.config.eab_strict_mode)
                mock_warning.assert_called_once()

    def test_013_parse_eab_strict_mode_none_uses_default(self):
        config_dic = {"EABhandler": {}}
        self.directory.config.eab_strict_mode = False

        result = self.directory._parse_eab_strict_mode(config_dic)

        self.assertFalse(result)

    def test_014_parse_eab_strict_mode_bool_raw_value(self):
        config_dic = {"EABhandler": {"eab_strict_mode": False}}
        self.directory.config.eab_strict_mode = True

        result = self.directory._parse_eab_strict_mode(config_dic)

        self.assertFalse(result)

    def test_015_parse_eab_strict_mode_eabhandler_access_error_uses_default(self):
        config_dic = MagicMock()
        config_dic.getboolean.side_effect = Exception("cannot parse bool")
        config_dic.__contains__.side_effect = Exception("contains failed")
        self.directory.config.eab_strict_mode = False

        result = self.directory._parse_eab_strict_mode(config_dic)

        self.assertFalse(result)

    def test_016_parse_eab_strict_mode_true_token_on(self):
        class ConfigLike(dict):
            def getboolean(self, *args, **kwargs):
                raise Exception("force raw parsing path")

        config_dic = ConfigLike({"EABhandler": {"eab_strict_mode": "on"}})
        self.directory.config.eab_strict_mode = False

        result = self.directory._parse_eab_strict_mode(config_dic)

        self.assertTrue(result)

    def test_017_load_ca_handler_success(self):
        config_dic = {}
        ca_handler_module = MagicMock()
        ca_handler_module.CAhandler = MagicMock()
        with patch(
            "acme2certifier.acme_srv.directory.ca_handler_load", return_value=ca_handler_module
        ):
            self.directory._load_ca_handler(config_dic)
            self.assertEqual(self.directory.cahandler, ca_handler_module.CAhandler)

    def test_018_load_ca_handler_failure(self):
        config_dic = {}
        with patch("acme2certifier.acme_srv.directory.ca_handler_load", return_value=None):
            with patch.object(self.mock_logger, "critical") as mock_critical:
                self.directory._load_ca_handler(config_dic)
                mock_critical.assert_called()

    def test_019_build_meta_information(self):
        self.directory.config.suppress_product_information = False
        self.directory.config.supress_version = False
        self.directory.config.tos_url = "tos"
        self.directory.config.caaidentities = ["id1"]
        self.directory.config.profiles = {"profile": "data"}
        self.directory.config.eab = True
        self.directory.config.eab_strict_mode = True
        meta = self.directory._build_meta_information()
        self.assertIn("home", meta)
        self.assertIn("author", meta)
        self.assertIn("name", meta)
        self.assertIn("version", meta)
        self.assertIn("termsOfService", meta)
        self.assertIn("caaIdentities", meta)
        self.assertIn("profiles", meta)
        self.assertIn("externalAccountRequired", meta)
        self.assertTrue(meta["externalAccountRequired"])

    def test_020_build_meta_information_eab_mixed_mode(self):
        self.directory.config.eab = True
        self.directory.config.eab_strict_mode = False
        meta = self.directory._build_meta_information()
        self.assertIn("externalAccountRequired", meta)
        self.assertFalse(meta["externalAccountRequired"])

    def test_021_build_meta_information_suppress(self):
        self.directory.config.suppress_product_information = True
        self.directory.config.home = "custom_home"
        meta = self.directory._build_meta_information()
        self.assertIn("home", meta)
        self.assertNotIn("author", meta)
        self.assertNotIn("name", meta)
        self.assertNotIn("version", meta)

    def test_022_build_directory_response_db_check_ok(self):
        self.directory.config.db_check = True
        self.directory.dbversion = "1.0"
        self.directory.repository = self.mock_repository
        with patch.object(
            self.mock_repository, "get_db_version", return_value=("1.0", "script")
        ):
            resp = self.directory._build_directory_response()
            self.assertEqual(resp["meta"]["db_check"], "OK")

    def test_023_build_directory_response_db_check_nok(self):
        self.directory.config.db_check = True
        self.directory.dbversion = "1.0"
        self.directory.repository = self.mock_repository
        with patch.object(
            self.mock_repository, "get_db_version", return_value=("2.0", "script")
        ):
            with patch.object(self.mock_logger, "error") as mock_error:
                resp = self.directory._build_directory_response()
                self.assertEqual(resp["meta"]["db_check"], "NOK")
                mock_error.assert_called()

    def test_024_build_directory_response_db_exception(self):
        self.directory.config.db_check = True
        self.directory.dbversion = "1.0"
        self.directory.repository = self.mock_repository
        with patch.object(
            self.mock_repository, "get_db_version", return_value=(None, None)
        ):
            with patch.object(self.mock_logger, "error") as mock_error:
                resp = self.directory._build_directory_response()
                self.assertEqual(resp["meta"]["db_check"], "NOK")
                mock_error.assert_called()

    def test_025_build_directory_response_random_key(self):
        resp = self.directory._build_directory_response()
        found_random = any(
            v
            == "https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417"
            for v in resp.values()
        )
        self.assertTrue(found_random)

    def test_026_get_directory_response_success(self):
        self.directory.cahandler = self.mock_cahandler
        self.mock_cahandler_instance.handler_check.return_value = None
        resp = self.directory.get_directory_response()
        self.assertIn("newAuthz", resp)
        self.assertIn("meta", resp)

    def test_027_get_directory_response_error(self):
        self.directory.cahandler = self.mock_cahandler
        self.mock_cahandler_instance.handler_check.return_value = "error"
        resp = self.directory.get_directory_response()
        self.assertIn("error", resp)

    def test_028_get_directory_response_no_handler(self):
        self.directory.cahandler = None
        resp = self.directory.get_directory_response()
        self.assertIn("error", resp)

    def test_029_directory_get(self):
        with patch.object(
            self.directory, "get_directory_response", return_value={"key": "value"}
        ):
            resp = self.directory.directory_get()
            self.assertEqual(resp, {"key": "value"})

    def test_030_servername_get(self):
        self.directory.server_name = "test_server"
        self.assertEqual(self.directory.servername_get(), "test_server")

    def test_031_parse_directory_section_calls_parse_caaidentities(self):
        # Mock config_dic to behave like configparser.ConfigParser
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "Directory"
        config_dic.__getitem__.side_effect = lambda k: (
            {"tos_url": "tos", "url_prefix": "/prefix", "home": "custom_home"}
            if k == "Directory"
            else None
        )
        # Return a non-None value for caaidentities
        config_dic.get.side_effect = lambda section, key, fallback=None: (
            '["id1", "id2"]' if key == "caaidentities" else None
        )
        with patch.object(
            self.directory,
            "_parse_caaidentities",
            wraps=self.directory._parse_caaidentities,
        ) as mock_parse_caaidentities:
            self.directory._parse_directory_section(config_dic)
            mock_parse_caaidentities.assert_called_once_with('["id1", "id2"]')

    def test_032_repository_get_db_version_success(self):
        mock_dbstore = MagicMock()
        mock_dbstore.dbversion_get.return_value = ("1.0", "script")
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        result = repo.get_db_version()
        self.assertEqual(result, ("1.0", "script"))

    def test_033_repository_get_db_version_exception(self):
        mock_dbstore = MagicMock()
        mock_dbstore.dbversion_get.side_effect = Exception("fail")
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        with patch.object(self.mock_logger, "critical") as mock_critical:
            result = repo.get_db_version()
            self.assertEqual(result, (None, None))
            mock_critical.assert_called()

    def test_034_get_directory_response_profiles_sync_load_profiles(self):
        # Setup Directory with profiles_sync enabled and no error from handler_check
        self.directory.config.profiles_sync = True
        self.directory.config.acme_url = "https://acme.example.com"
        self.directory.config.profiles_sync_interval = 1234
        self.directory.config.async_mode = False
        self.directory.config.profiles = {}
        mock_cahandler_instance = MagicMock()
        mock_cahandler_instance.__enter__.return_value = mock_cahandler_instance
        mock_cahandler_instance.__exit__.return_value = None
        mock_cahandler_instance.handler_check.return_value = None
        mock_cahandler_instance.synchronize_profiles.return_value = {
            "profile": "loaded"
        }
        self.directory.cahandler = MagicMock(return_value=mock_cahandler_instance)
        # Ensure hasattr returns True for load_profiles
        with patch.object(
            mock_cahandler_instance,
            "load_profiles",
            wraps=mock_cahandler_instance.synchronize_profiles,
        ):
            resp = self.directory.get_directory_response()
            self.assertEqual(self.directory.config.profiles, {"profile": "loaded"})
            self.assertIn("newAuthz", resp)

    def test_035_get_directory_response_no_cahandler(self):
        self.directory.cahandler = None
        with patch.object(self.mock_logger, "critical") as mock_critical:
            resp = self.directory.get_directory_response()
            self.assertIn("error", resp)
            mock_critical.assert_called()

    def test_036_profile_list_get_success(self):
        mock_dbstore = MagicMock()
        profiles_json = None
        mock_dbstore.hkparameter_get.return_value = profiles_json
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        self.assertFalse(repo.profile_list_get())

    def test_037_profile_list_get_success(self):
        mock_dbstore = MagicMock()
        profiles_json = '[{"name": "profile1"}, {"name": "profile2"}]'
        mock_dbstore.hkparameter_get.return_value = profiles_json
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        result = repo.profile_list_get()
        self.assertEqual(result, [{"name": "profile1"}, {"name": "profile2"}])

    def test_038_profile_list_get_db_exception(self):
        mock_dbstore = MagicMock()
        mock_dbstore.hkparameter_get.side_effect = Exception("fail")
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        with patch.object(self.mock_logger, "critical") as mock_critical:
            result = repo.profile_list_get()
            self.assertEqual(result, [])
            mock_critical.assert_called()

    def test_039_profile_list_get_json_error(self):
        mock_dbstore = MagicMock()
        # Use an invalid JSON string to ensure json.loads fails
        mock_dbstore.hkparameter_get.return_value = "{invalid_json: true]"
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        with patch.object(self.mock_logger, "error") as mock_error:
            result = repo.profile_list_get()
            self.assertEqual(result, [])
            mock_error.assert_called()

    def test_040_profile_list_set_success(self):
        mock_dbstore = MagicMock()
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        data_dic = {"profiles": ["profile1", "profile2"]}
        repo.profile_list_set(data_dic)
        mock_dbstore.hkparameter_add.assert_called_once_with(data_dic)

    def test_041_profile_list_set_db_exception(self):
        mock_dbstore = MagicMock()
        mock_dbstore.hkparameter_add.side_effect = Exception("fail")
        repo = DirectoryRepository(mock_dbstore, self.mock_logger)
        data_dic = {"profiles": ["profile1", "profile2"]}
        with patch.object(self.mock_logger, "critical") as mock_critical:
            repo.profile_list_set(data_dic)
            mock_critical.assert_called()

    def test_042_parse_cahandler_section_profiles_sync_exception(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        config_dic.getboolean.side_effect = Exception("fail")
        with patch.object(self.mock_logger, "error") as mock_error:
            self.directory._parse_cahandler_section(config_dic)
            mock_error.assert_any_call("profiles_sync not set: %s", ANY)

    def test_043_parse_cahandler_section_sets_acme_url(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: False
        self.directory._parse_cahandler_section(config_dic)
        self.assertEqual(self.directory.config.acme_url, "https://acme.example.com")

    def test_044_parse_cahandler_section_profiles_sync_disabled(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: False
        self.directory.config.profiles = {}
        self.directory._parse_cahandler_section(config_dic)
        self.assertFalse(self.directory.config.profiles_sync)

    def test_045_parse_cahandler_section_profiles_sync_enabled_profiles_configured(
        self,
    ):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        # First call to getboolean returns True for profiles_sync
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: (
            True if key == "profiles_sync" else False
        )
        self.directory.config.profiles = {"profile": "data"}
        with patch.object(self.mock_logger, "error") as mock_error:
            self.directory._parse_cahandler_section(config_dic)
            self.assertFalse(self.directory.config.profiles_sync)
            mock_error.assert_any_call(
                "Profiles are configured via acme_srv.cfg. Disabling profile sync."
            )

    def test_046_parse_cahandler_section_profiles_sync_enabled_no_acme_url(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: {} if k == "CAhandler" else None
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: (
            True if key == "profiles_sync" else False
        )
        self.directory.config.profiles = {}
        self.directory.config.acme_url = None
        with patch.object(self.mock_logger, "error") as mock_error:
            self.directory._parse_cahandler_section(config_dic)
            self.assertFalse(self.directory.config.profiles_sync)
            mock_error.assert_any_call(
                "profiles_sync is set but no acme_url configured."
            )

    def test_047_parse_cahandler_section_profiles_sync_interval_set(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: (
            True if key == "profiles_sync" else False
        )
        config_dic.getint.side_effect = lambda section, key, fallback=None: (
            1234 if key == "profiles_sync_interval" else fallback
        )
        self.directory.config.profiles = {}
        self.directory.config.acme_url = "https://acme.example.com"
        self.directory.config.profiles_sync = False
        self.directory._parse_cahandler_section(config_dic)
        self.assertEqual(self.directory.config.profiles_sync_interval, 1234)

    def test_048_parse_cahandler_section_profiles_sync_interval_error(self):
        config_dic = MagicMock()
        config_dic.__contains__.side_effect = lambda k: k == "CAhandler"
        config_dic.__getitem__.side_effect = lambda k: (
            {"acme_url": "https://acme.example.com"} if k == "CAhandler" else None
        )
        config_dic.getboolean.side_effect = lambda section, key, fallback=None: (
            True if key == "profiles_sync" else False
        )
        config_dic.getint.side_effect = Exception("fail")
        self.directory.config.profiles = {}
        self.directory.config.acme_url = "https://acme.example.com"
        self.directory.config.profiles_sync = False
        with patch.object(self.mock_logger, "error") as mock_error:
            self.directory._parse_cahandler_section(config_dic)
            mock_error.assert_any_call("profiles_sync_interval not set: %s", ANY)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Cryptographic operations for acme2certifier"""

import hashlib
import json
import logging
import random
import re
from string import digits, ascii_letters
from typing import Dict, Tuple
from jwcrypto import jwk, jws
from .encoding import b64decode_pad, b64_encode
from .validation import dkeys_lower


def decode_deserialize(logger: logging.Logger, string: str) -> Dict:
    """decode and deserialize string"""
    logger.debug("Helper.decode_deserialize()")
    # b64 decode
    string_decode = b64decode_pad(logger, string)
    # deserialize if b64 decoding was successful
    if string_decode and string_decode != "ERR: b64 decoding error":
        try:
            string_decode = json.loads(string_decode)
        except ValueError:
            string_decode = "ERR: Json decoding error"

    return string_decode


def decode_message(
    logger: logging.Logger, message: str
) -> Tuple[str, str, Dict[str, str], Dict[str, str], str]:
    """decode jwstoken and return header, payload and signature"""
    logger.debug("Helper.decode_message()")
    jwstoken = jws.JWS()
    result = False
    error = None
    try:
        jwstoken.deserialize(message)
        protected = json.loads(jwstoken.objects["protected"])
        if bool(jwstoken.objects["payload"]):
            payload = json.loads(jwstoken.objects["payload"])
        else:
            payload = {}
        signature = jwstoken.objects["signature"]
        result = True
    except Exception as err:
        logger.error("Error during message decoding %s", err)
        error = str(err)
        protected = {}
        payload = {}
        signature = None

    if payload:
        payload = dkeys_lower(payload)
    return (result, error, protected, payload, signature)


def generate_random_string(logger: logging.Logger, length: int) -> str:
    """generate random string to be used as name"""
    logger.debug("Helper.generate_random_string()")
    char_set = digits + ascii_letters
    return "".join(random.choice(char_set) for _ in range(length))


def jwk_thumbprint_get(logger: logging.Logger, pub_key: Dict[str, str]) -> str:
    """get thumbprint"""
    logger.debug("Helper.jwk_thumbprint_get()")
    if pub_key:
        try:
            jwkey = jwk.JWK(**pub_key)
            thumbprint = jwkey.thumbprint()
        except Exception as err:
            logger.error("Could not get the JWKEY thumbprint from public key: %s", err)
            thumbprint = None
    else:
        thumbprint = None

    logger.debug("Helper.jwk_thumbprint_get() ended with: %s", thumbprint)
    return thumbprint


def sha256_hash(logger: logging.Logger, string: str) -> str:
    """hash string"""
    logger.debug("Helper.sha256_hash()")
    result = hashlib.sha256(string.encode("utf-8")).digest()
    logger.debug(
        "Helper.sha256_hash() ended with %s (base64-encoded)",
        b64_encode(logger, result),
    )
    return result


def sha256_hash_hex(logger: logging.Logger, string: str) -> str:
    """hash string"""
    logger.debug("Helper.sha256_hash_hex()")
    result = hashlib.sha256(string.encode("utf-8")).hexdigest()
    logger.debug("Helper.sha256_hash_hex() ended with %s", result)
    return result


def signature_check(
    logger: logging.Logger, message: str, pub_key: str, json_: bool = False
) -> Tuple[bool, str]:
    """check JWS"""
    logger.debug("Helper.signature_check(%s)", json_)

    result = False
    error = None

    if pub_key:
        # load key
        try:
            if json_:
                logger.debug("Helper.signature_check(): load key from json")
                jwkey = jwk.JWK.from_json(pub_key)
            else:
                logger.debug("Helper.signature_check(): load plain json")
                jwkey = jwk.JWK(**pub_key)
        except Exception as err:
            logger.error("Loading of public key failed %s", err)
            jwkey = None
            result = False
            error = str(err)

        # verify signature
        if jwkey:
            jwstoken = jws.JWS()
            jwstoken.deserialize(message)
            try:
                jwstoken.verify(jwkey)
                result = True
            except Exception as err:
                logger.error("Message verification failed %s", err)
                error = str(err)
        else:
            logger.error("No jwkey extracted")
    else:
        logger.error("No pubkey specified.")
        error = "No key specified."

    logger.debug("Helper.signature_check() ended with: %s, %s", result, error)
    return (result, error)


def string_sanitize(logger: logging.Logger, unsafe_str: str) -> str:
    """sanitize string"""
    logger.debug("Helper.string_sanitize()")
    allowed_range = set(range(32, 127))
    safe_str = ""
    for char in unsafe_str:
        cp_ = ord(char)
        if cp_ in allowed_range:
            safe_str += char
        elif cp_ == 9:
            safe_str += " " * 4
    return re.sub(r"\s+", " ", safe_str)



"""
Email Reply Challenge Validator.

Implements validation logic for email-reply-00 challenges.
"""

from typing import Optional, Tuple
from .base import ChallengeValidator, ChallengeContext, ValidationResult
from acme2certifier.acme_srv.helper import (
    b64_url_encode,
    convert_byte_to_string,
    sha256_hash,
)


class EmailReplyChallengeValidator(ChallengeValidator):
    """Validator for email-reply-00 challenges."""

    def get_challenge_type(self) -> str:
        """Return the challenge type this validator handles."""
        return "email-reply-00"

    def perform_validation(self, context: ChallengeContext) -> ValidationResult:
        """Perform email-reply-00 challenge validation."""
        self.logger.debug("EmailReplyChallengeValidator.perform_validation()")
        try:
            from acme2certifier.acme_srv.email_handler import EmailHandler
        except ImportError as e:
            return ValidationResult(
                success=False,
                invalid=True,
                error_message=f"Email handler not available: {e}",
                details={"import_error": str(e)},
            )

        calculated_keyauth, rfc_token1 = self._generate_email_keyauth(
            context.challenge_name,
            context.token,
            context.jwk_thumbprint,
            context.keyauthorization,
        )

        with EmailHandler(debug=False, logger=self.logger) as email_handler:
            email_receive = email_handler.receive(
                callback=lambda email_data: self._filter_email(email_data, rfc_token1)
            )

            if not email_receive or "body" not in email_receive:
                return ValidationResult(
                    success=False,
                    invalid=False,
                    error_message="No email received or email body missing",
                )

            email_keyauth = self._extract_email_keyauth(email_receive["body"])

            if (
                email_keyauth
                and calculated_keyauth
                and email_keyauth == calculated_keyauth
            ):
                self.logger.debug(
                    "EmailReplyChallengeValidator.perform_validation() complete"
                )
                return ValidationResult(
                    success=True,
                    invalid=False,
                    details={"calculated_keyauth": calculated_keyauth},
                )
            else:
                self.logger.error(
                    "Email keyauthorization does not match calculated keyauthorization"
                )
                return ValidationResult(
                    success=False,
                    invalid=True,
                    error_message="Email keyauthorization mismatch",
                    details={"expected": calculated_keyauth, "received": email_keyauth},
                )

    def _generate_email_keyauth(
        self, challenge_name: str, rfc_token2: str, jwk_thumbprint: str, rfc_token1: str
    ) -> Tuple[str, str]:
        """Generate email keyauthorization - placeholder for actual implementation."""
        self.logger.debug(
            "EmailReplyChallengeValidator._generate_email_keyauth() for %s",
            challenge_name,
        )

        calculated_keyauth = convert_byte_to_string(
            b64_url_encode(
                self.logger,
                sha256_hash(self.logger, f"{rfc_token1}{rfc_token2}.{jwk_thumbprint}"),
            )
        )
        return calculated_keyauth, rfc_token1

    def _filter_email(self, email_data, rfc_token1):

        filter_string = f"ACME: {rfc_token1}"
        self.logger.debug(
            "Challenge._validate_email_reply_challenge(): filter string: %s",
            filter_string,
        )

        if filter_string in email_data.get("subject", ""):
            self.logger.debug(
                "Challenge._validate_email_reply_challenge(): email subject matches filter: %s",
                email_data["subject"],
            )
            return email_data
        else:
            self.logger.debug(
                "Challenge._validate_email_reply_challenge(): email subject does not match filter: %s",
                email_data.get("subject", ""),
            )
            return None

    def _extract_email_keyauth(self, email_body: Optional[str]) -> Optional[str]:
        """Extract keyauthorization from email body between PEM-style delimiters."""
        self.logger.debug("EmailReplyChallengeValidator._extract_email_keyauth()")
        begin = "-----BEGIN ACME RESPONSE-----"
        end = "-----END ACME RESPONSE-----"
        email_keyauthorization = None
        if email_body:
            start = email_body.find(begin)
            if start >= 0:
                start += len(begin)
                stop = email_body.find(end, start)
                if stop >= 0:
                    email_keyauthorization = email_body[start:stop].strip() or None

        self.logger.debug(
            "Challenge._emailchallenge_keyauth_extract() ended with: %s",
            bool(email_keyauthorization),
        )
        return email_keyauthorization



# -*- coding: utf-8 -*-
"""urls for acme django database"""

from django.urls import re_path
from acme2certifier.django_app import views

urlpatterns = [
    re_path(r"^acct", views.acct, name="acct"),
    re_path(r"^authz", views.authz, name="authz"),
    re_path(r"^cert", views.cert, name="cert"),
    re_path(r"^chall", views.chall, name="chall"),
    re_path(r"^directory$", views.directory, name="directory"),
    re_path(r"^newaccount$", views.newaccount, name="newaccount"),
    re_path(r"^key-change$", views.acct, name="acct"),
    re_path(r"^newnonce$", views.newnonce, name="newnonce"),
    re_path(r"^neworders$", views.neworders, name="neworders"),
    re_path(r"^order", views.order, name="order"),
    re_path(r"^revokecert", views.revokecert, name="revokecert"),
    re_path(r"^renewal-info", views.renewalinfo, name="renewalinfo"),
    re_path(r"^servername_get$", views.servername_get, name="servername_get"),
]



# -*- coding: utf-8 -*-
# pylint: disable=w0611
"""
Backwards compatibility layer for acme2certifier helper functions.
This file imports all functions from the modular helpers structure
to maintain compatibility with existing code.
"""

# Encoding and base64 operations
from .helpers.encoding import (
    b64decode_pad,
    b64_decode,
    b64_encode,
    b64_url_encode,
    b64_url_recode,
    b64_url_decode,
    build_pem_file,
    convert_byte_to_string,
    convert_string_to_byte,
)

# Certificate operations
from .helpers.certificates import (
    cert_aki_get,
    cert_load,
    cert_dates_get,
    cert_cn_get,
    cert_der2pem,
    cert_issuer_get,
    cert_pem2der,
    cert_pubkey_get,
    cert_san_get,
    cert_ski_get,
    cryptography_version_get,
    cert_extensions_get,
    cert_serial_get,
    pembundle_to_list,
    certid_asn1_get,
    certid_hex_get,
    certid_check,
    pkcs7_to_pem,
)

# CSR operations
from .helpers.csr import (
    csr_load,
    csr_cn_get,
    csr_dn_get,
    csr_pubkey_get,
    csr_san_get,
    csr_san_byte_get,
    csr_extensions_get,
    csr_subject_get,
    csr_cn_lookup,
)

# Cryptographic operations
from .helpers.crypto import (
    decode_deserialize,
    decode_message,
    generate_random_string,
    jwk_thumbprint_get,
    sha256_hash,
    sha256_hash_hex,
    signature_check,
    string_sanitize,
)

# Date/time utilities
from .helpers.datetime_utils import (
    uts_now,
    uts_to_date_utc,
    date_to_uts_utc,
    date_to_datestr,
    datestr_to_date,
)

# Validation functions
from .helpers.validation import (
    dkeys_lower,
    fqdn_in_san_check,
    validate_csr,
    validate_email,
    validate_identifier,
    validate_ip,
    validate_fqdn,
    ip_validate,
    ipv6_chk,
    cn_validate,
)

# Network operations
from .helpers.network import (
    _fqdn_resolve,
    fqdn_resolve,
    ptr_resolve,
    dns_server_list_load,
    patched_create_connection,
    proxy_check,
    url_get_with_own_dns,
    allowed_gai_family,
    url_get_with_default_dns,
    url_get,
    txt_get,
    proxystring_convert,
    servercert_get,
    v6_adjust,
    header_info_get,
    get_url,
    parse_url,
    encode_url,
    request_operation,
)

# Configuration
from .helpers.config import (
    config_check,
    config_profile_load,
    config_eab_profile_load,
    config_headerinfo_load,
    config_enroll_config_log_load,
    config_allowed_domainlist_load,
    config_allowed_iplist_load,
    config_async_mode_load,
    config_proxy_load,
    config_dns_server_list_load,
    config_dryrun_load,
    load_config,
    header_info_jsonify,
    header_info_lookup,
    client_parameter_validate,
    profile_lookup,
)

# Logging utilities
from .helpers.logging_utils import (
    _logger_nonce_modify,
    _logger_certificate_modify,
    _logger_token_modify,
    _logger_challenges_modify,
    logger_info,
    logger_setup,
    print_debug,
    handle_exception,
)

# Plugin loaders
from .helpers.plugin_loader import (
    ca_handler_load,
    db_handler_load,
    eab_handler_load,
    hooks_load,
)

# EAB functions
from .helpers.eab import (
    eab_profile_header_info_check,
    eab_profile_subject_string_check,
    eab_profile_subject_check,
    eab_profile_revocation_check,
    eab_profile_check,
    eab_profile_list_check,
    eab_profile_string_check,
)

# Domain utilities
from .helpers.domain_utils import (
    encode_domain,
    wildcard_domain_check,
    pattern_check,
    is_domain_whitelisted,
    allowed_domainlist_check,
    sancheck_lists_create,
    is_ip_whitelisted,
)

# General utilities
from .helpers.utils import (
    error_dic_get,
    enrollment_config_log,
    radomize_parameter_list,
    handler_config_check,
)



# -*- coding: utf-8 -*-
"""Refactored Account class with improved design and maintainability"""

from __future__ import print_function
import json
from typing import List, Tuple, Dict, Optional
from dataclasses import dataclass, field
from acme2certifier.acme_srv.helper import (
    generate_random_string,
    validate_email,
    date_to_datestr,
    load_config,
    eab_handler_load,
    b64decode_pad,
    error_dic_get,
    uts_to_date_utc,
    uts_now,
)
from acme2certifier.acme_srv.db_handler import DBstore
from acme2certifier.acme_srv.message import Message

from acme2certifier.acme_srv.signature import Signature

DB_ERROR_MSG = "Database error"


class ExternalAccountBinding:
    """Encapsulates EAB validation and signature verification logic."""

    def __init__(self, logger, eab_handler, server_name=None):
        self.logger = logger
        self.eab_handler = eab_handler
        self.server_name = server_name

    def get_kid(self, protected: str) -> str:
        """Extract key identifier (kid) from protected header."""
        self.logger.debug("ExternalAccountBinding.get_kid()")
        try:
            protected_dic = json.loads(b64decode_pad(self.logger, protected))
        except Exception as err:
            self.logger.error("Failed to decode protected header: %s", err)
            protected_dic = None

        if isinstance(protected_dic, dict):
            eab_key_id = protected_dic.get("kid", None)
        else:
            eab_key_id = None
        self.logger.debug("ExternalAccountBinding.get_kid() ended with: %s", eab_key_id)
        return eab_key_id

    def compare_jwk(self, protected: dict, payload: str) -> bool:
        """Compare JWK from outer header with JWK in EAB payload."""
        self.logger.debug("ExternalAccountBinding.compare_jwk()")
        result = False
        if "jwk" in protected:
            jwk_outer = protected["jwk"]
            try:
                jwk_inner = json.loads(b64decode_pad(self.logger, payload))
            except Exception as err:
                self.logger.error("Failed to decode EAB JWK payload: %s", err)
            else:
                if isinstance(jwk_inner, dict):
                    if json.dumps(jwk_outer, sort_keys=True) == json.dumps(
                        jwk_inner, sort_keys=True
                    ):
                        result = True
                    else:
                        self.logger.error("JWK from outer and inner JWS do not match")
                        self.logger.debug("outer: %s", jwk_outer)
                        self.logger.debug("inner: %s", jwk_inner)
                else:
                    self.logger.error("EAB JWK payload is not a JSON object")
        else:
            self.logger.error("No JWK in protected header")
        self.logger.debug("ExternalAccountBinding.compare_jwk() ended with: %s", result)
        return result

    def verify_signature(self, content: dict, mac_key: str) -> tuple:
        """Verify EAB signature."""
        self.logger.debug("ExternalAccountBinding.verify_signature()")
        if content and mac_key:
            signature = Signature(None, self.server_name, self.logger)
            jwk_ = json.dumps({"k": mac_key, "kty": "oct"})
            sig_check, error = signature.eab_check(json.dumps(content), jwk_)
        else:
            sig_check = False
            error = None
        self.logger.debug(
            "ExternalAccountBinding.verify_signature() ended with: %s: %s",
            sig_check,
            error,
        )
        return (sig_check, error)

    def verify(self, payload: dict, err_msg_dic: dict) -> tuple:
        """Check for external account binding and verify signature."""
        self.logger.debug("ExternalAccountBinding.verify()")
        eab_kid = self.get_kid(payload["externalaccountbinding"]["protected"])
        if eab_kid:
            with self.eab_handler(self.logger) as eab_handler:
                eab_mac_key = eab_handler.mac_key_get(eab_kid)
        else:
            eab_mac_key = None
        if eab_mac_key:
            result, error = self.verify_signature(
                payload["externalaccountbinding"], eab_mac_key
            )
            if result:
                code = 200
                message = None
                detail = None
            else:
                code = 403
                message = err_msg_dic["unauthorized"]
                detail = "EAB signature verification failed"
                self.logger.error("EAB verification returned an error: %s", error)
        else:
            code = 403
            message = err_msg_dic["unauthorized"]
            detail = "EAB kid lookup failed"
        self.logger.debug("ExternalAccountBinding.verify() ended with: %s", code)
        return (code, message, detail)

    def check(self, protected: dict, payload: dict, err_msg_dic: dict) -> tuple:
        """Check for external account binding, compare JWK, and verify signature."""
        self.logger.debug("ExternalAccountBinding.check()")

        if (
            self.eab_handler
            and protected
            and payload
            and "externalaccountbinding" in payload
            and payload["externalaccountbinding"]
        ):
            jwk_compare = self.compare_jwk(
                protected, payload["externalaccountbinding"]["payload"]
            )
            if jwk_compare and "protected" in payload["externalaccountbinding"]:
                return self.verify(payload, err_msg_dic)
            else:
                code = 403
                message = err_msg_dic["malformed"]
                detail = "Malformed request"
        else:
            code = 403
            message = err_msg_dic["externalaccountrequired"]
            detail = "External account binding required"
        self.logger.debug("ExternalAccountBinding.check() ended with: %s", code)
        return (code, message, detail)


class AccountDatabaseError(Exception):
    """Exception raised for database-related errors in Account operations."""

    pass


class AccountRepository:
    """Repository for all Account-related database operations."""

    def __init__(self, dbstore, logger=None):
        self.dbstore = dbstore
        self.logger = logger

    def lookup_account(self, field: str, value: str) -> Optional[Dict[str, str]]:
        """Look up an account in the database."""
        try:
            return self.dbstore.account_lookup(field, value)
        except Exception as err:
            self.logger.critical("Database error during account lookup: %s", err)
            raise AccountDatabaseError(f"Failed to look up account: {err}") from err

    def add_account(self, data_dic: Dict[str, str]) -> Tuple[Optional[str], bool]:
        """Add a new account to the database."""
        try:
            return self.dbstore.account_add(data_dic)
        except Exception as err:
            self.logger.critical("Database error while adding account: %s", err)
            raise AccountDatabaseError(f"Failed to add account: {err}") from err

    def update_account(self, data_dic: Dict[str, str], active: bool = True) -> bool:
        """Update an account in the database."""
        try:
            return self.dbstore.account_update(data_dic, active)
        except Exception as err:
            self.logger.critical("Database error while updating account: %s", err)
            raise AccountDatabaseError(f"Failed to update account: {err}") from err

    def delete_account(self, account_name: str) -> bool:
        """Delete an account from the database."""
        try:
            return self.dbstore.account_delete(account_name)
        except Exception as err:
            self.logger.critical("Database error while deleting account: %s", err)
            raise AccountDatabaseError(f"Failed to delete account: {err}") from err

    def load_jwk(self, account_name: str) -> Optional[Dict[str, str]]:
        """Load the JWK for a given account."""
        try:
            return self.dbstore.jwk_load(account_name)
        except Exception as err:
            self.logger.critical("Database error while loading JWK: %s", err)
            raise AccountDatabaseError(f"Failed to load JWK: {err}") from err


@dataclass
class AccountConfiguration:
    """Configuration for the Account class."""

    ecc_only: bool = False
    contact_check_disable: bool = False
    tos_check_disable: bool = False
    inner_header_nonce_allow: bool = False
    tos_url: Optional[str] = None
    eab_check: bool = False
    eab_strict_mode: bool = True
    eab_handler: Optional[object] = None
    path_dic: Dict[str, str] = field(
        default_factory=lambda: {"acct_path": "/acme/acct/"}
    )


@dataclass
class AccountData:
    """Data structure for account information."""

    name: str
    alg: str
    jwk: Dict[str, str]
    contact: List[str]
    eab_kid: Optional[str] = None
    status: str = "valid"
    created_at: Optional[str] = None


class Account:
    """Refactored ACME server class."""

    def __init__(self, debug: bool = False, srv_name: str = None, logger=None):
        self.server_name = srv_name
        self.logger = logger
        self.dbstore = DBstore(debug, self.logger)
        self.repository = AccountRepository(self.dbstore, self.logger)
        self.message = Message(debug, self.server_name, self.logger)
        self.config = AccountConfiguration()
        self.err_msg_dic = error_dic_get(self.logger)

    def __enter__(self):
        """Enter the context manager, loading configuration."""
        self._load_configuration()
        return self

    def __exit__(self, *args) -> None:
        """
        Exit the context manager. (No-op, placeholder for cleanup.)
        """

    def _load_configuration(self):
        """Load configuration into the AccountConfiguration dataclass."""
        self.logger.debug("Account._load_configuration()")
        config_dic = load_config()

        self.config.inner_header_nonce_allow = config_dic.getboolean(
            "Account", "inner_header_nonce_allow", fallback=False
        )
        self.config.ecc_only = config_dic.getboolean(
            "Account", "ecc_only", fallback=False
        )
        self.config.tos_check_disable = config_dic.getboolean(
            "Account", "tos_check_disable", fallback=False
        )
        self.config.contact_check_disable = config_dic.getboolean(
            "Account", "contact_check_disable", fallback=False
        )

        if "EABhandler" in config_dic:
            self.logger.debug("Account._load_configuration(): loading eab_handler")
            self.config.eab_check = True

            self.config.eab_strict_mode = config_dic.getboolean(
                "EABhandler", "eab_strict_mode", fallback=True
            )
            if (
                "eab_handler_file" in config_dic["EABhandler"]
                or "eab_handler_module" in config_dic["EABhandler"]
            ):
                eab_handler_module = eab_handler_load(self.logger, config_dic)
                if eab_handler_module:
                    self.config.eab_handler = eab_handler_module.EABhandler
                else:
                    self.logger.critical("EABHandler could not get loaded")
            else:
                self.logger.critical("EABHandler configuration incomplete")

        self.config.tos_url = config_dic.get("Directory", "tos_url", fallback=None)
        if config_dic.get("Directory", "url_prefix", fallback=None):
            self.config.path_dic = {
                k: config_dic.get("Directory", "url_prefix") + v
                for k, v in self.config.path_dic.items()
            }
        self.logger.debug("Account._load_configuration() ended")

    def _add_account_to_db(
        self, account_data: AccountData
    ) -> Tuple[int, str, Optional[Dict[str, str]]]:
        """Add a new account to the database."""
        self.logger.debug("Account._add_account_to_db(%s)", account_data.name)
        try:
            # convert dict and list to string
            account_data.jwk = json.dumps(account_data.jwk)
            account_data.contact = json.dumps(account_data.contact)
            db_name, is_new = self.repository.add_account(account_data.__dict__)
            if is_new:
                self.logger.debug(
                    "Account._add_account_to_db() ended with: 201, %s", db_name
                )
                return 201, db_name, None
            self.logger.debug(
                "Account._add_account_to_db() ended with: 200, %s", db_name
            )
            return 200, db_name, None

        except Exception as err:
            self.logger.critical("Database error while adding account: %s", err)
            return 500, self.err_msg_dic["serverinternal"], DB_ERROR_MSG

    def _validate_contact(self, contact: List[str]) -> Tuple[int, str, str]:
        """Validate contact information."""
        self.logger.debug("Account._validate_contact()")
        if not contact:
            return 400, self.err_msg_dic["malformed"], "Contact information is missing"
        if not validate_email(self.logger, contact):
            return (
                400,
                self.err_msg_dic["invalidcontact"],
                "Invalid contact information",
            )
        return 200, None, None

    def _check_tos(self, content: Dict[str, str]) -> Tuple[int, str, str]:
        """check terms of service"""
        self.logger.debug("Account._check_tos()")
        if "termsofserviceagreed" in content:
            self.logger.debug("tos:%s", content["termsofserviceagreed"])
            if content["termsofserviceagreed"]:
                code = 200
                message = None
                detail = None
            else:
                code = 403
                message = self.err_msg_dic["useractionrequired"]
                detail = "Terms of service must be agreed"
        else:
            self.logger.debug("no tos statement found.")
            code = 403
            message = self.err_msg_dic["useractionrequired"]
            detail = "termsofserviceagreed flag missing"

        self.logger.debug("Account._check_tos() ended with:%s", code)
        return (code, message, detail)

    def _skip_eab_for_non_strict_mode(self, payload: Dict[str, str]) -> bool:
        """Return True if EAB checks should be skipped in non-strict mode."""
        return not self.config.eab_strict_mode and not payload.get(
            "externalaccountbinding"
        )

    def _run_eab_check(
        self, protected: Dict[str, str], payload: Dict[str, str]
    ) -> Tuple[int, Optional[str], Optional[str]]:
        """Run EAB validation when required by current configuration."""
        if not self.config.eab_check:
            return 200, None, None

        if self._skip_eab_for_non_strict_mode(payload):
            self.logger.debug(
                "Account._create_account() EAB check skipped due to non-strict mode and missing EAB payload"
            )
            return 200, None, None

        eab_handler = ExternalAccountBinding(
            self.logger, self.config.eab_handler, self.server_name
        )
        return eab_handler.check(protected, payload, self.err_msg_dic)

    def _get_eab_kid_for_account(
        self, account_name: str, payload: Dict[str, str]
    ) -> Optional[str]:
        """Extract EAB kid for account data when EAB data is provided/required."""
        if not self.config.eab_check:
            return None

        if self._skip_eab_for_non_strict_mode(payload):
            self.logger.debug(
                "Account._create_account() EAB binding skipped for account %s due to non-strict mode and missing EAB payload",
                account_name,
            )
            return None

        eab_handler = ExternalAccountBinding(
            self.logger, self.config.eab_handler, self.server_name
        )
        return eab_handler.get_kid(payload["externalaccountbinding"]["protected"])

    def _create_account(
        self, payload: Dict[str, str], protected: Dict[str, str]
    ) -> Tuple[int, str, str]:
        """Create a new account."""
        self.logger.debug("Account._create_account()")
        account_name = generate_random_string(self.logger, 12)
        contact_list = payload.get("contact", [])

        # tos check
        if self.config.tos_url and not self.config.tos_check_disable:
            code, message, detail = self._check_tos(payload)
            if code != 200:
                return code, message, detail

        code, message, detail = self._run_eab_check(protected, payload)
        if code != 200:
            return code, message, detail

        # Validate contact information
        if not self.config.contact_check_disable:
            code, message, detail = self._validate_contact(contact_list)
            if code != 200:
                return code, message, detail

        # Prepare account data
        account_data = AccountData(
            name=account_name,
            alg=protected["alg"],
            jwk=protected["jwk"],
            contact=contact_list,
            created_at=date_to_datestr(uts_now()),
        )
        eab_kid = self._get_eab_kid_for_account(account_name, payload)
        if eab_kid:
            account_data.eab_kid = eab_kid

        # Add account to database
        return self._add_account_to_db(account_data)

    def _parse_query(self, account_name: str) -> Dict[str, str]:
        """update contacts"""
        self.logger.debug("Account._parse_query(%s)", account_name)

        # this is a query for account information
        account_obj = self._lookup_account_by_name(account_name)
        if account_obj:
            data = self._build_account_info(account_obj)
            data["status"] = "valid"
        else:
            data = {"status": "invalid"}

        self.logger.debug("Account._parse_query() ended")
        return data

    def _onlyreturnexisting(
        self, protected: Dict[str, str], payload: Dict[str, str]
    ) -> Tuple[int, str, str]:
        """check onlyreturnexisting"""
        self.logger.debug("Account._onlyreturnexisting()")

        if "onlyreturnexisting" in payload:
            if payload["onlyreturnexisting"]:

                if "jwk" in protected:
                    result = self._lookup_account_by_field(
                        json.dumps(protected["jwk"]), "jwk"
                    )
                    if result:
                        code = 200
                        message = result["name"]
                        detail = self._parse_query(message)
                    else:
                        code = 400
                        message = self.err_msg_dic["accountdoesnotexist"]
                        detail = None
                else:
                    code = 400
                    message = self.err_msg_dic["malformed"]
                    detail = "jwk structure missing"

            else:
                code = 400
                message = self.err_msg_dic["useractionrequired"]
                detail = "onlyReturnExisting must be true"
        else:
            code = 500
            message = self.err_msg_dic["serverinternal"]
            detail = "onlyReturnExisting without payload"

        self.logger.debug("Account.onlyreturnexisting() ended with: %s", code)
        return (code, message, detail)

    def _handle_deactivation(
        self, account_name: str, payload: Dict[str, str]
    ) -> Dict[str, str]:
        """Handle account deactivation."""
        self.logger.debug("Account._handle_deactivation(%s)", account_name)
        if payload.get("status", "").lower() == "deactivated":
            account_obj = self._lookup_account_by_name(account_name)
            if not account_obj:
                return self._build_response(
                    400,
                    self.err_msg_dic["accountdoesnotexist"],
                    "Deactivation failed",
                )
            code, message, detail = self._deactivate_account(account_name)
            if code == 200:
                data = self._build_account_info(account_obj)
                data["status"] = "deactivated"
                return self._build_response(code, account_name, data)
            return self._build_response(code, message, detail)
        else:
            return self._build_response(
                400, self.err_msg_dic["malformed"], "Invalid status for deactivation"
            )

    def _deactivate_account(self, account_name: str) -> Tuple[int, str, str]:
        """Deactivate an account."""
        self.logger.debug("Account._deactivate_account(%s)", account_name)
        try:
            data_dic = {
                "name": account_name,
                "status_id": 7,
                "jwk": f"DEACTIVATED {uts_to_date_utc(uts_now())}",
            }
            result = self.repository.update_account(data_dic, active=False)
            if result:
                return 200, None, None
            else:
                return (
                    400,
                    self.err_msg_dic["accountdoesnotexist"],
                    "Deactivation failed",
                )
        except Exception as err:
            self.logger.critical("Database error while deactivating account: %s", err)
            return 500, self.err_msg_dic["serverinternal"], DB_ERROR_MSG

    def _handle_contact_update(
        self, account_name: str, payload: Dict[str, str]
    ) -> Dict[str, str]:
        """Handle contact update for an account."""
        self.logger.debug("Account._handle_contact_update(%s)", account_name)
        code, message, detail = self._update_account_contacts(account_name, payload)
        if code == 200:
            account_obj = self._lookup_account_by_name(account_name)
            if account_obj:
                data = self._build_account_info(account_obj)
                return self._build_response(code, message, data)
        return self._build_response(code, message, detail)

    def _update_account_contacts(
        self, account_name: str, payload: Dict[str, str]
    ) -> Tuple[int, str, str]:
        """Update account contacts in the database."""
        self.logger.debug("Account._update_account_contacts(%s)", account_name)
        code, message, detail = self._validate_contact(payload.get("contact", []))
        if code != 200:
            return code, message, detail

        try:
            data_dic = {"name": account_name, "contact": json.dumps(payload["contact"])}
            result = self.repository.update_account(data_dic)
            if result:
                return 200, None, None
            else:
                return 400, self.err_msg_dic["accountdoesnotexist"], "Update failed"
        except Exception as err:
            self.logger.critical(
                "Database error while updating account contacts: %s", err
            )
            return 500, self.err_msg_dic["serverinternal"], DB_ERROR_MSG

    def _handle_key_change(
        self, account_name: str, payload: Dict[str, str], protected: Dict[str, str]
    ) -> Dict[str, str]:
        """Handle key change for an account."""
        self.logger.debug("Account._handle_key_change(%s)", account_name)
        if "url" in protected and "key-change" in protected["url"]:
            (
                code,
                _message,
                _detail,
                inner_protected,
                inner_payload,
                _,
            ) = self.message.check(
                json.dumps(payload), use_emb_key=True, skip_nonce_check=True
            )
            if code == 200:
                code, message, _detail = self._rollover_account_key(
                    account_name, protected, inner_protected, inner_payload
                )
                if code == 200:
                    return self._build_response(code, message, None)
        return self._build_response(
            400, self.err_msg_dic["malformed"], "Malformed key-change request"
        )

    def _rollover_account_key(
        self,
        account_name: str,
        protected: Dict[str, str],
        inner_protected: Dict[str, str],
        inner_payload: Dict[str, str],
    ) -> Tuple[int, str, str]:
        """Perform key rollover for an account."""
        self.logger.debug("Account._rollover_account_key(%s)", account_name)
        code, message, detail = self._validate_key_change(
            account_name, protected, inner_protected, inner_payload
        )
        if code == 200:
            try:
                data_dic = {
                    "name": account_name,
                    "jwk": json.dumps(inner_protected["jwk"]),
                }
                result = self.repository.update_account(data_dic)
                if result:
                    return 200, None, None
                else:
                    self.logger.error(
                        "Key rollover failed for account: %s", account_name
                    )
                    return (
                        500,
                        self.err_msg_dic["serverinternal"],
                        "Key rollover failed",
                    )
            except Exception as err:
                self.logger.critical(
                    "Database error while updating account key: %s", err
                )
                return 500, self.err_msg_dic["serverinternal"], DB_ERROR_MSG
        return code, message, detail

    def _validate_key_change(
        self,
        account_name: str,
        protected: Dict[str, str],
        inner_protected: Dict[str, str],
        inner_payload: Dict[str, str],
    ) -> Tuple[int, str, str]:
        """Validate key change request."""
        self.logger.debug("Account._validate_key_change(%s)", account_name)
        if "jwk" not in inner_protected:
            return 400, self.err_msg_dic["malformed"], "Inner JWS is missing JWK"

        key_exists = self._lookup_account_by_field(
            json.dumps(inner_protected["jwk"]), "jwk"
        )
        if key_exists:
            return 400, self.err_msg_dic["badpubkey"], "Public key already exists"

        if "url" in protected and "url" in inner_protected:
            if protected["url"] != inner_protected["url"]:
                return (
                    400,
                    self.err_msg_dic["malformed"],
                    "URL mismatch in inner and outer JWS",
                )
        else:
            return (
                400,
                self.err_msg_dic["malformed"],
                "Missing URL in inner or outer JWS",
            )

        if "kid" in protected and "account" in inner_payload:
            if protected["kid"] != inner_payload["account"]:
                return (
                    400,
                    self.err_msg_dic["malformed"],
                    "KID and account do not match",
                )
        else:
            return (
                400,
                self.err_msg_dic["malformed"],
                "Missing KID or account in payload",
            )

        return 200, None, None

    def _handle_account_query(self, account_name: str) -> Dict[str, str]:
        """Handle account query."""
        self.logger.debug("Account._handle_account_query(%s)", account_name)
        account_obj = self._lookup_account_by_name(account_name)
        if account_obj:
            data = self._build_account_info(account_obj)
            return self._build_response(200, account_name, data)
        return self._build_response(
            400, self.err_msg_dic["accountdoesnotexist"], "Account not found"
        )

    def _lookup_account_by_name(self, value: str) -> Optional[Dict[str, str]]:
        """Lookup an account in the database."""
        self.logger.debug("Account._lookup_account_by_name(name: %s)", value)
        try:
            return self.repository.lookup_account("name", value)
        except Exception as err:
            self.logger.critical(
                "Database error during account lookup by name: %s", err
            )
            return None

    def _lookup_account_by_field(
        self, value: str, field: str
    ) -> Optional[Dict[str, str]]:
        """Lookup account by a specific field."""
        self.logger.debug("Account._lookup_account_by_field(%s: %s)", field, value)
        try:
            return self.repository.lookup_account(field, value)
        except Exception as err:
            self.logger.critical(
                "Database error during account lookup by %s: %s", field, err
            )
            return None

    def _build_account_info(self, account_obj: Dict[str, str]) -> Dict[str, str]:
        """Build account information for response."""
        self.logger.debug("Account._build_account_info()")
        account_info = {
            "status": account_obj.get("status", "valid"),
            "key": json.loads(account_obj["jwk"]),
            "contact": json.loads(account_obj["contact"]),
            "createdAt": date_to_datestr(account_obj["created_at"]),
        }
        if "eab_kid" in account_obj and account_obj["eab_kid"]:
            account_info["eab_kid"] = account_obj["eab_kid"]

        self.logger.debug(
            "Account._build_account_info() ended with: %s", bool(account_info)
        )
        return account_info

    def _build_response(
        self,
        code: int,
        message: Optional[str],
        detail: Optional[str],
        payload: Optional[Dict] = None,
    ) -> Dict[str, str]:
        """Build a response dictionary."""
        self.logger.debug("Account._build_response()")
        response_dic = {}
        if code in (200, 201):
            response_dic["data"] = {}
            if code == 201:
                response_dic["data"] = {
                    "status": "valid",
                    "orders": f'{self.server_name}{self.config.path_dic["acct_path"]}{message}/orders',
                }
                if payload and "contact" in payload:
                    response_dic["data"]["contact"] = payload["contact"]
            elif code == 200 and detail and "status" in detail:
                response_dic["data"] = detail

            response_dic["header"] = {}
            response_dic["header"][
                "Location"
            ] = f'{self.server_name}{self.config.path_dic["acct_path"]}{message}'

            # add exernal account binding
            if (
                self.config.eab_check
                and payload
                and "externalaccountbinding" in payload
            ):
                response_dic["data"]["externalaccountbinding"] = payload[
                    "externalaccountbinding"
                ]

        else:
            if detail == "tosfalse":
                detail = "Terms of service must be accepted"

        # prepare/enrich response
        status_dic = {"code": code, "type": message, "detail": detail}
        response_dic = self.message.prepare_response(response_dic, status_dic)

        return response_dic

    def create_account(self, content: Dict[str, str]) -> Dict[str, str]:
        """Public method to create a new account."""
        self.logger.debug("Account.create_account()")
        code, message, detail, protected, payload, _ = self.message.check(content, True)
        if code != 200:
            return self._build_response(code, message, detail, payload)

        # onlyReturnExisting check
        if "onlyreturnexisting" in payload:
            code, message, detail = self._onlyreturnexisting(protected, payload)
        else:
            code, message, detail = self._create_account(payload, protected)

        self.logger.debug("Account.create_account() ended with: %s, %s", code, message)
        return self._build_response(code, message, detail, payload)

    def parse_request(self, content: Dict[str, str]) -> Dict[str, str]:
        """Public method to parse an account-related request."""
        self.logger.debug("Account.parse_request()")
        code, message, detail, protected, payload, account_name = self.message.check(
            content
        )
        if code != 200:
            return self._build_response(code, message, detail)

        if "status" in payload:
            return self._handle_deactivation(account_name, payload)
        elif "contact" in payload:
            return self._handle_contact_update(account_name, payload)
        elif "payload" in payload:
            return self._handle_key_change(account_name, payload, protected)
        elif not payload:
            return self._handle_account_query(account_name)
        else:
            return self._build_response(
                400, self.err_msg_dic["malformed"], "Unknown request"
            )

    # Compatibility layer for external methods
    def new(self, content: Dict[str, str]) -> Dict[str, str]:
        """Compatibility layer for the new method."""
        return self.create_account(content)

    def parse(self, content: Dict[str, str]) -> Dict[str, str]:
        """Compatibility layer for the parse method."""
        return self.parse_request(content)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""eab file handler"""

from __future__ import print_function
from typing import Dict
import csv

# pylint: disable=E0401
from acme2certifier.acme_srv.helper import load_config


class EABhandler(object):
    """EAB file handler"""

    def __init__(self, logger: object = None):
        self.logger = logger
        self.key_file = None

    def __enter__(self):
        """Makes EABhandler a Context Manager"""
        if not self.key_file:
            self._config_load()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _config_load(self):
        """ " load config from file"""
        self.logger.debug("EABhandler._config_load()")

        config_dic = load_config(self.logger, "EABhandler")
        self.key_file = config_dic.get("EABhandler", "key_file", fallback=self.key_file)

        self.logger.debug("EABhandler._config_load() ended")

    def key_file_load(self) -> Dict[str, str]:
        """load key_file"""
        self.logger.debug("EABhandler.key_file_load()")

        data_dic = {}
        if self.key_file:
            try:
                with open(self.key_file, mode="r", encoding="utf8") as csv_file:
                    csv_reader = csv.DictReader(csv_file)
                    for row in csv_reader:
                        data_dic[row["eab_kid"]] = row["eab_mac"]
            except Exception as err:
                self.logger.error("Failed to load EAB key file: %s", err)

        self.logger.debug("EABhandler.key_file_load() ended: {%s}", bool(data_dic))
        return data_dic

    def mac_key_get(self, kid: str = None) -> str:
        """check external account binding"""
        self.logger.debug("EABhandler.mac_key_get(%s)", kid)

        mac_key = None
        if self.key_file and kid:
            data_dic = self.key_file_load()
            if kid in data_dic:
                mac_key = data_dic[kid]
            else:
                self.logger.error(
                    "MAC key retrieval failed: kid '%s' not found in key file.", kid
                )
        self.logger.debug("EABhandler.mac_key_get() ended with: %s", bool(mac_key))
        return mac_key



# -*- coding: utf-8 -*-
# pylint: disable=C0302
"""Dogtag CA handler"""

from __future__ import print_function
from typing import Tuple, Dict, Optional
import os
import requests
from requests_pkcs12 import Pkcs12Adapter
from cryptography import x509

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    load_config,
    b64_decode,
    b64_encode,
    cert_pem2der,
    cert_serial_get,
    handler_config_check,
    eab_profile_header_info_check,
    config_enroll_config_log_load,
    config_profile_load,
    config_eab_profile_load,
    config_headerinfo_load,
    pkcs7_to_pem,
    enrollment_config_log,
    uts_now,
    uts_to_date_utc,
)


def update_validity_attributes(logger, data: dict, notbefore: str, notafter: str):
    """
    Update notBefore and notAfter attribute values in the profile policy set.

    Args:
        logger (logging.Logger): Logger instance for debug output.
        data (dict): Profile policy set data structure.
        notbefore (str): NotBefore date string.
        notafter (str): NotAfter date string.
    Returns:
        None
    """
    logger.debug("update_validity_attributes()")
    for policy in data.get("ProfilePolicySet", [])[0].get("policies", []):
        for attr in policy.get("def", {}).get("attributes", []):
            if attr.get("name") == "notBefore":
                # Do not log actual notBefore value (could be sensitive)
                logger.debug(
                    "update_validity_attributes(): Setting notBefore attribute value (redacted)"
                )
                attr["Value"] = notbefore
            elif attr.get("name") == "notAfter":
                # Do not log actual notAfter value (could be sensitive)
                logger.debug(
                    "update_validity_attributes(): Setting notAfter attribute value (redacted)"
                )
                attr["Value"] = notafter
    logger.debug("update_validity_attributes() ended")


def approve_profile_get(logger, nonce: str = None, request_id: str = None):
    """
    Build the approval profile data structure for a certificate request.

    Args:
        logger (logging.Logger): Logger instance for debug output.
        nonce (Optional[str]): Nonce value for the request.
        request_id (Optional[str]): Certificate request ID.
    Returns:
        dict: Approval profile data structure.
    """
    # Do not log nonce (sensitive); log only request_id
    logger.debug("approve_profile_get() called with request_id: %s", request_id)

    notbefore = uts_to_date_utc(uts_now() - 300, tformat="%Y-%m-%d %H:%M:%S")
    notafter = uts_to_date_utc(uts_now() + 365 * 24 * 3600, tformat="%Y-%m-%d %H:%M:%S")

    data = {
        "nonce": None,
        "requestId": None,
        "ProfilePolicySet": [
            {
                "policies": [
                    {
                        "id": "2",
                        "def": {
                            "name": "Validity Default",
                            "text": "This default populates a Certificate Validity to the request. The default values are Range=720 in days",
                            "attributes": [
                                {
                                    "name": "notBefore",
                                    "Value": None,
                                    "Descriptor": {
                                        "Syntax": "string",
                                        "Description": "Not Before",
                                    },
                                },
                                {
                                    "name": "notAfter",
                                    "Value": None,
                                    "Descriptor": {
                                        "Syntax": "string",
                                        "Description": "Not After",
                                    },
                                },
                            ],
                        },
                    },
                    {
                        "id": "5",
                        "def": {
                            "name": "AIA Extension Default",
                            "attributes": [
                                {
                                    "name": "authInfoAccessCritical",
                                    "Value": "false",
                                    "Descriptor": {
                                        "Syntax": "boolean",
                                        "Description": "Criticality",
                                        "DefaultValue": "false",
                                    },
                                },
                                {
                                    "name": "authInfoAccessGeneralNames",
                                    "Value": "Record #0\r\nMethod:1.3.6.1.5.5.7.48.1\r\nLocation Type:URIName\r\nLocation:http://dogtag.acme:8080/ca/ocsp\r\nEnable:true\r\n\r\n",
                                    "Descriptor": {
                                        "Syntax": "string_list",
                                        "Description": "General Names",
                                    },
                                },
                            ],
                        },
                    },
                    {
                        "id": "6",
                        "def": {
                            "name": "Key Usage Default",
                            "attributes": [
                                {
                                    "name": "keyUsageCritical",
                                    "Value": "true",
                                    "Descriptor": {
                                        "Syntax": "boolean",
                                        "Description": "Criticality",
                                        "DefaultValue": "false",
                                    },
                                },
                                {
                                    "name": "keyUsageDigitalSignature",
                                    "Value": "true",
                                    "Descriptor": {
                                        "Syntax": "boolean",
                                        "Description": "Digital Signature",
                                        "DefaultValue": "false",
                                    },
                                },
                                {
                                    "name": "keyUsageKeyEncipherment",
                                    "Value": "true",
                                    "Descriptor": {
                                        "Syntax": "boolean",
                                        "Description": "Key Encipherment",
                                        "DefaultValue": "false",
                                    },
                                },
                                {
                                    "name": "keyUsageDataEncipherment",
                                    "Value": "true",
                                    "Descriptor": {
                                        "Syntax": "boolean",
                                        "Description": "Data Encipherment",
                                        "DefaultValue": "false",
                                    },
                                },
                            ],
                        },
                        "constraint": {
                            "name": "Key Usage Extension Constraint",
                            "text": "This constraint accepts the Key Usage extension, if present, only when Criticality=true, Digital Signature=true, Non-Repudiation=false, Key Encipherment=true, Data Encipherment=true, Key Agreement=false, Key Certificate Sign=false, Key CRL Sign=false, Encipher Only=false, Decipher Only=false",
                            "classId": "KeyUsageExtConstraint",
                        },
                    },
                    {
                        "id": "7",
                        "def": {
                            "name": "Extended Key Usage Extension Default",
                            "attributes": [
                                {
                                    "name": "exKeyUsageCritical",
                                    "Value": "false",
                                },
                                {
                                    "name": "exKeyUsageOIDs",
                                    "Value": "1.3.6.1.5.5.7.3.1,1.3.6.1.5.5.7.3.2",
                                },
                            ],
                        },
                    },
                    {
                        "id": "8",
                        "def": {
                            "name": "Signing Alg",
                            "attributes": [
                                {
                                    "name": "signingAlg",
                                    "Value": "SHA256withRSA",
                                }
                            ],
                        },
                    },
                ]
            }
        ],
        "Attributes": {"Attribute": []},
    }

    data["requestId"] = request_id
    data["nonce"] = nonce

    # update notBefore and notAfter values in the profile policy set
    update_validity_attributes(logger, data, notbefore, notafter)

    logger.debug("approve_profile_get() ended")
    return data


class CAhandler(object):
    """Dogtag CA handler"""

    # Class-level constants for magic strings and numbers
    DEFAULT_PREFIX = "/ca"
    REST_AGENT_CERTREQUESTS = "/ca/rest/agent/certrequests/"
    REST_CERTREQUESTS = "/ca/rest/certrequests"
    REST_AGENT_CERTS = "/ca/rest/agent/certs/"
    REST_CERTS = "/ca/rest/certs/"
    REST_INFO = "/pki/rest/info"
    DEFAULT_REQUEST_TIMEOUT = 30
    CONFIG_SECTION = "CAhandler"

    def __init__(self, _debug: Optional[bool] = None, logger: Optional[object] = None):
        """
        Initialize the CAhandler.

        Args:
            _debug (Optional[bool]): Enable debug mode. Not currently used.
            logger (Optional[logging.Logger]): Logger instance for logging.
        """
        self.logger = logger
        self.client_cert = None
        self.client_key = None
        self.cert_passphrase = None
        self.api_host = None
        self.api_version = None
        self.ca_bundle = None
        self.certrequest_approve = False
        self.profile = None
        self.session = None
        self.header_info_field = False
        self.eab_handler = None
        self.eab_profiling = False
        self.enrollment_config_log = False
        self.enrollment_config_log_skip_list = []
        self.profiles = {}
        self.request_timeout = self.DEFAULT_REQUEST_TIMEOUT
        self.proxy = None
        self.profile_mapping_field = "profile"

    def __enter__(self) -> "CAhandler":
        """
        Enter the context manager, loading configuration and logging in if needed.

        Returns:
            CAhandler: The handler instance.
        """
        if not self.api_host:
            self._config_load()
            self._login()
        return self

    def __exit__(self, *args) -> None:
        """
        Close the connection at the end of the context.

        Args:
            *args: Exception type, value, and traceback (if any).
        Returns:
            None
        """
        if self.session:
            self.session.close()
            self.session = None

    def _api_post(self, url: str, data: Dict[str, str]) -> Tuple[int, Dict[str, str]]:
        """
        Generic wrapper for an API POST call.

        Args:
            url (str): API endpoint URL (relative).
            data (Dict[str, str]): Data to send in the POST request.
        Returns:
            Tuple[int, Dict[str, str]]: HTTP status code and response content.
        """
        self.logger.debug("_api_post(%s)", url)

        if self.session is None:
            self._login()

        try:
            response = self.session.post(
                url=self.api_host + url,
                json=data,
                proxies=self.proxy,
                verify=self.ca_bundle,
                timeout=self.request_timeout,
            )
            code = response.status_code
            try:
                content = response.json()
            except Exception as err_:
                self.logger.error(
                    "Could not parse the response for an API post() request: %s", err_
                )
                content = {"error": f"Could not parse JSON: {err_}"}
        except Exception as err_:
            self.logger.error("API post() returned error: %s", err_)
            code = 500
            content = {"error": str(err_)}

        return code, content

    def _api_get(self, url: str) -> Tuple[int, Dict[str, str]]:
        """
        Get data from API via GET request.

        Args:
            url (str): API endpoint URL (relative).
        Returns:
            Tuple[int, Dict[str, str]]: HTTP status code and response content.
        """
        self.logger.debug("CAhandler._api_get()")
        headers = None

        if self.session is None:
            self._login()

        try:
            api_response = self.session.get(
                url=self.api_host + url,
                headers=headers,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            )
            code = api_response.status_code
            try:
                content = api_response.json()
            except Exception as err_:
                self.logger.error(
                    "Could not parse the response for an API get() request: %s", err_
                )
                content = {"error": f"Could not parse JSON: {err_}"}
        except Exception as err_:
            self.logger.error("API get() request returned error: %s", err_)
            code = 500
            content = {"error": str(err_)}

        return code, content

    def _api_version_get(self) -> None:
        """
        Get API version from CA and set self.api_version.
        Returns:
            None
        """
        self.logger.debug("CAhandler._api_version_get()")

        if not self.api_version:
            code, content = self._api_get(self.REST_INFO)

            if code == 200 and isinstance(content, dict):
                self.api_version = next(
                    (v for k, v in content.items() if k.lower() == "version"), "unknown"
                )
                self.logger.info("CA API version: %s", self.api_version)
            else:
                self.logger.error(
                    "Failed to get CA API version. Status code: %s, Response: %s",
                    code,
                    content,
                )
                self.api_version = "unknown"

    def _get_approval_nonce(
        self, request_id: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Get nonce for certificate request approval.

        Args:
            request_id (str): Certificate request ID.
        Returns:
            Tuple[Optional[str], Optional[str]]: Error message (or None), nonce value (or None).
        """
        self.logger.debug("CAhandler._get_approval_nonce(%s)", request_id)

        nonce = None
        error = None

        # we need a nonce first to lets send a get to the request status endpoint to get the nonce
        code, response = self._api_get(f"{self.REST_AGENT_CERTREQUESTS}{request_id}")

        if "requestStatus" in response:
            # Do not log full status if it contains sensitive info
            self.logger.debug(
                "CAHandler._certrequest_approve() - Certificate request %s status before approval attempt",
                request_id,
            )

        if code == 200 and isinstance(response, dict) and "nonce" in response:
            nonce = response["nonce"]
            # Do not log nonce value (sensitive)
            self.logger.debug("Received nonce for approval (redacted)")

        else:
            self.logger.error(
                "Failed to get nonce for certificate request approval. Status code: %s, Response: %s",
                code,
                response,
            )
            error = "Failed to get nonce for certificate request approval."

        self.logger.debug(
            "CAhandler._get_approval_nonce() ended with error: %s",
            error if error else "None",
        )
        return error, nonce

    def _certrequest_approve(
        self, request_id: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Approve certificate request with given request_id.

        Args:
            request_id (str): Certificate request ID.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str]]: Error message (or None), cert_bundle, cert_raw.
        """
        self.logger.debug("CAhandler._certrequest_approve(%s)", request_id)
        error, nonce = self._get_approval_nonce(request_id)
        if error:
            self.logger.error(
                "Cannot approve certificate request %s due to error: %s",
                request_id,
                error,
            )
            return (error, None, None)

        # Approve request
        data = approve_profile_get(self.logger, nonce, request_id)
        code, response = self._api_post(
            f"{self.REST_AGENT_CERTREQUESTS}{request_id}/approve", data
        )

        if code >= 400:
            return self._approve_error(
                "Failed to approve certificate request.",
                code,
                response,
            )

        # Get new status and certificate data
        code, response = self._api_get(f"{self.REST_CERTREQUESTS}/{request_id}")
        if "requestStatus" not in response:
            return self._approve_error(
                "Failed to approve certificate request or retrieve certificate data after approval attempt.",
                code,
                response,
            )
        if "certId" not in response:
            return self._approve_error(
                "Certificate request approved but certId is missing in the response, cannot fetch certificate data.",
                code,
                response,
            )

        self.logger.debug(
            "CACertificate request %s approved successfully (certId redacted)",
            request_id,
        )
        return self._fetch_cert_and_bundle(response["certId"])

    def _fetch_cert_and_bundle(
        self, cert_id: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Fetch certificate and bundle by cert_id.

        Args:
            cert_id (str): Certificate ID.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str]]: Error message (or None), cert_bundle, cert_raw.
        """
        code, response = self._api_get(f"{self.REST_CERTS}{cert_id}")
        error = None
        cert_raw = None
        cert_bundle = None
        if code == 200:
            if "Encoded" in response and response["Encoded"]:
                cert_raw = b64_encode(
                    self.logger, cert_pem2der(response.get("Encoded"))
                )
            else:
                self.logger.error(
                    "Certificate request approved but failed to fetch raw certificate data. Status code: %s",
                    code,
                )
                error = "Certificate request approved but failed to fetch raw certificate data."
            if "PKCS7CertChain" in response and response["PKCS7CertChain"]:
                cert_bundle = pkcs7_to_pem(
                    self.logger, b64_decode(self.logger, response.get("PKCS7CertChain"))
                )
            else:
                self.logger.error(
                    "Certificate request approved but failed to fetch certificate chain data. Status code: %s",
                    code,
                )
                error = "Certificate request approved but failed to fetch certificate chain data."
        else:
            error = f"Certificate request approved but failed to fetch certificate data. Status code: {code}, Response: {response}"
        self.logger.debug(
            "CAhandler._certrequest_approve() ended with error: %s", error
        )
        return (error, cert_bundle, cert_raw)

    def _approve_error(self, msg: str, code: int, response) -> Tuple[str, None, None]:
        """
        Log and return error for approval failures.

        Args:
            msg (str): Error message.
            code (int): HTTP status code.
            response: Response object or dict.
        Returns:
            Tuple[str, None, None]: Error message, None, None.
        """
        # For compatibility with tests, always include the legacy substring in the log
        legacy_msg = "Failed to get certificate request status after approval attempt"
        self.logger.error(
            f"{legacy_msg}. Status code: %s, Response: %s", code, response
        )
        self.logger.debug("CAhandler._certrequest_approve() ended with error: %s", msg)
        return (msg, None, None)

    def _parse_cert(self, pem_data: str) -> None:
        """
        Parse a PEM-encoded certificate and print key usage and extended key usage.

        Args:
            pem_data (str): PEM-encoded certificate string.
        Returns:
            None
        """
        cert = x509.load_pem_x509_certificate(pem_data.encode())

        try:
            ku_ext = cert.extensions.get_extension_for_class(x509.KeyUsage)
            print("\nKey Usage:")
            ku_value = ku_ext.value
            # Einzelne Flags prüfen (z.B. digital_signature, key_encipherment, etc.)
            usages = [
                "digital_signature" if ku_value.digital_signature else None,
                "content_commitment" if ku_value.content_commitment else None,
                "key_encipherment" if ku_value.key_encipherment else None,
                "data_encipherment" if ku_value.data_encipherment else None,
                "key_agreement" if ku_value.key_agreement else None,
                "key_cert_sign" if ku_value.key_cert_sign else None,
                "crl_sign" if ku_value.crl_sign else None,
            ]
            print(f"  Active: {[u for u in usages if u]}")
        except x509.ExtensionNotFound:
            print("No Key Usage extension found.")

        try:
            eku_ext = cert.extensions.get_extension_for_class(x509.ExtendedKeyUsage)
            print("Extended Key Usage (OIDs):")
            for oid in eku_ext.value:
                # pylint: disable=protected-access
                print(f"  - {oid._name} ({oid.dotted_string})")
        except x509.ExtensionNotFound:
            print("No Extended Key Usage extension found.")

    def _certrequest_send(
        self, csr: Optional[str] = None
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Send a certificate signing request (CSR) to the CA.

        Args:
            csr (Optional[str]): PEM-encoded CSR string.
        Returns:
            Tuple[Optional[str], Optional[str]]: Request ID and request status.
        """
        self.logger.debug("CAhandler._certrequest_send()")

        data = {
            "ProfileID": self.profile,
            "Input": [
                {
                    "id": "i1",
                    "ClassID": "certReqInputImpl",
                    "Name": "Certificate Request Input",
                    "ConfigAttribute": [],
                    "Attribute": [
                        {
                            "name": "cert_request_type",
                            "Value": "pkcs10",
                            "Descriptor": {
                                "Syntax": "cert_request_type",
                                "Description": "Certificate Request Type",
                            },
                        },
                        {
                            "name": "cert_request",
                            "Value": csr,
                            "Descriptor": {
                                "Syntax": "cert_request",
                                "Description": "Certificate Request",
                            },
                        },
                    ],
                }
            ],
        }

        url = self.REST_CERTREQUESTS
        code, response = self._api_post(url, data)

        request_id = None
        request_status = None
        if code == 200:
            if isinstance(response, dict) and "entries" in response:
                self.logger.info("Certificate request sent successfully")
                # Process the response as needed
                if (
                    response["entries"]
                    and "requestId" in response["entries"][0]
                    and "requestStatus" in response["entries"][0]
                ):
                    request_id = response["entries"][0]["requestId"]
                    request_status = response["entries"][0]["requestStatus"]
                    self.logger.info(
                        "Request ID: %s, Request Status: %s", request_id, request_status
                    )

            else:
                self.logger.error(
                    "Unexpected response format for certificate request. Response: %s",
                    response,
                )

        else:
            self.logger.error(
                "Failed to send certificate request. Status code: %s, Response: %s",
                code,
                response,
            )

        self.logger.debug(
            "CAhandler._certrequest_send() ended with code: %s, request_id: %s, request_status: %s",
            code,
            request_id,
            request_status,
        )

        return request_id, request_status

    def _config_passphrase_load(self, config_dic) -> None:
        """
        Load certificate passphrase from environment variable or config.

        Args:
            config_dic: Configuration dictionary or ConfigParser.
        Returns:
            None
        """
        section = None
        # Support both dict and ConfigParser
        if isinstance(config_dic, dict) and self.CONFIG_SECTION in config_dic:
            section = config_dic[self.CONFIG_SECTION]
        elif hasattr(config_dic, "items") and config_dic.has_section(
            self.CONFIG_SECTION
        ):
            # Convert section to dict
            section = dict(config_dic.items(self.CONFIG_SECTION))
        if not isinstance(section, dict):
            return

        if "cert_passphrase_variable" in section:
            self._load_passphrase_from_env(section["cert_passphrase_variable"])

        if "cert_passphrase" in section:
            self._load_passphrase_from_config(section["cert_passphrase"])

    def _load_passphrase_from_env(self, var_name: str) -> None:
        """
        Load certificate passphrase from environment variable.

        Args:
            var_name (str): Environment variable name.
        Returns:
            None
        """
        try:
            self.cert_passphrase = os.environ[var_name]
        except Exception as err:
            self.logger.error("Could not load cert_passphrase_variable:%s", err)

    def _load_passphrase_from_config(self, passphrase: str) -> None:
        """
        Load certificate passphrase from config value.

        Args:
            passphrase (str): Passphrase value.
        Returns:
            None
        """
        if self.cert_passphrase:
            self.logger.info("CAhandler._config_load() overwrite cert_passphrase")
        self.cert_passphrase = passphrase

    def _config_load(self) -> None:
        """
        Load configuration from file and set handler attributes.
        Returns:
            None
        """
        self.logger.debug("CAhandler._config_load()")

        config_dic = load_config(self.logger, self.CONFIG_SECTION)

        self.api_host = config_dic.get(
            self.CONFIG_SECTION, "api_host", fallback=self.api_host
        )

        self.client_cert = config_dic.get(
            self.CONFIG_SECTION, "client_cert", fallback=self.client_cert
        )

        self.client_key = config_dic.get(
            self.CONFIG_SECTION, "client_key", fallback=self.client_key
        )

        self.profile = config_dic.get(
            self.CONFIG_SECTION, self.profile_mapping_field, fallback=self.profile
        )

        self.ca_bundle = config_dic.get(
            self.CONFIG_SECTION, "ca_bundle", fallback=self.ca_bundle
        )

        if str(self.ca_bundle).lower() in ["true", "false"]:
            self.ca_bundle = config_dic.getboolean(
                self.CONFIG_SECTION, "ca_bundle", fallback=self.ca_bundle
            )

        self.certrequest_approve = config_dic.getboolean(
            self.CONFIG_SECTION,
            "certrequest_approve",
            fallback=self.certrequest_approve,
        )

        # load profiling
        self.eab_profiling, self.eab_handler = config_eab_profile_load(
            self.logger, config_dic
        )

        self._config_passphrase_load(config_dic)

        # load profiles
        self.profiles = config_profile_load(self.logger, config_dic)
        # load header info
        self.header_info_field = config_headerinfo_load(self.logger, config_dic)
        # load enrollment config log
        (
            self.enrollment_config_log,
            self.enrollment_config_log_skip_list,
        ) = config_enroll_config_log_load(self.logger, config_dic)

        self.logger.debug("CAhandler._config_load() ended")

    def _login(self) -> None:
        """
        Login to the CA API and establish a session.
        Returns:
            None
        """
        self.logger.debug("CAhandler._login()")

        if not self.session:
            self.session = requests.Session()

            if self.client_cert and self.cert_passphrase:
                self.logger.debug(
                    "CAhandler._login() using PKCS12 client authentication"
                )
                self.session.mount(
                    self.api_host,
                    Pkcs12Adapter(
                        pkcs12_filename=self.client_cert,
                        pkcs12_password=self.cert_passphrase,
                    ),
                )
            else:
                # client auth via pem files
                self.session.cert = (self.client_cert, self.client_key)
            # update hader
            self.session.headers.update({"Accept": "application/json"})

    def _revoke(self, serial: str) -> Tuple[int, str, str]:
        """
        Revoke certificate with given serial number.

        Args:
            serial (str): Certificate serial number.
        Returns:
            Tuple[int, str, str]: Code, message, and detail.
        """
        self.logger.debug("CAhandler._revoke(%s)", serial)
        serial = self._normalize_serial(serial)
        code, response = self._api_get(f"{self.REST_AGENT_CERTS}{serial}")
        if code != 200 or not isinstance(response, dict):
            return self._revoke_status_error(serial, code, response)
        if "Nonce" not in response:
            return self._revoke_status_error(serial, code, response)
        nonce = response["Nonce"]
        self.logger.debug(
            "CAhandler._revoke(): received nonce for revocation: %s", nonce
        )
        status = response.get("Status", "").lower()
        if status == "valid":
            return self._revoke_valid(serial, nonce)
        if status == "revoked":
            self.logger.info("Certificate with serial %s is already revoked", serial)
            return (200, "Certificate is already revoked", "")
        return self._revoke_status_error(serial, code, response)

    def _normalize_serial(self, serial: str) -> str:
        """
        Ensure serial starts with '0x'.

        Args:
            serial (str): Certificate serial number.
        Returns:
            str: Normalized serial number.
        """
        return serial if str(serial).lower().startswith("0x") else f"0x{serial}"

    def _revoke_valid(self, serial: str, nonce: str) -> Tuple[int, str, str]:
        """
        Handle revocation for valid certificates.

        Args:
            serial (str): Certificate serial number.
            nonce (str): Nonce value for revocation.
        Returns:
            Tuple[int, str, str]: Code, message, and detail.
        """
        self.logger.debug(
            "CAhandler._revoke(): Certificate with serial %s is valid, proceeding with revocation",
            serial,
        )
        data = {"Reason": "Unspecified", "Nonce": nonce}
        code, response = self._api_post(f"{self.REST_AGENT_CERTS}{serial}/revoke", data)
        if code == 200:
            self.logger.info("Certificate with serial %s revoked successfully", serial)
            return (code, "Certificate revoked successfully", "")
        self.logger.error(
            "Failed to revoke certificate with serial %s. Status code: %s, Response: %s",
            serial,
            code,
            response,
        )
        return (code, "Failed to revoke certificate", str(response))

    def _revoke_status_error(
        self, serial: str, code: int, response
    ) -> Tuple[int, str, str]:
        """
        Handle error when failing to get certificate status before revocation.

        Args:
            serial (str): Certificate serial number.
            code (int): HTTP status code.
            response: Response object or dict.
        Returns:
            Tuple[int, str, str]: Code, message, and detail.
        """
        self.logger.error(
            "Failed to get certificate status for serial %s before revocation attempt. Status code: %s, Response: %s",
            serial,
            code,
            response,
        )
        return (
            code,
            "Failed to get certificate status before revocation attempt",
            str(response),
        )

    def enroll(
        self, csr: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        """
        Enroll a certificate.

        Args:
            csr (str): CSR string.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]: Error, PEM bundle, raw certificate, poll identifier.
        """
        self.logger.debug("CAhandler.enroll()")

        cert_bundle = None
        cert_raw = None
        poll_identifier = None

        error: Optional[str] = eab_profile_header_info_check(
            self.logger, self, csr, self.profile_mapping_field
        )

        if error:
            self.logger.debug(
                "Certificate.enroll() ended() with error from profile header info check"
            )
            return (error, cert_bundle, cert_raw, poll_identifier)

        self._api_version_get()

        if self.enrollment_config_log:
            enrollment_config_log(
                self.logger, self, self.enrollment_config_log_skip_list
            )

        request_id, request_status = self._certrequest_send(csr)
        error, cert_bundle, cert_raw, poll_identifier = self._handle_enroll_status(
            request_id, request_status
        )

        self.logger.debug("Certificate.enroll() ended()")
        return (error, cert_bundle, cert_raw, poll_identifier)

    def _handle_enroll_status(
        self, request_id: Optional[str], request_status: Optional[str]
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        """
        Handle the status of the enrollment request.

        Args:
            request_id (Optional[str]): The request ID returned from the CA.
            request_status (Optional[str]): The status of the request.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]: Error, PEM bundle, raw certificate, poll identifier.
        """
        cert_bundle = None
        cert_raw = None
        poll_identifier = None
        error = None

        if request_status == "pending":
            if self.certrequest_approve:
                error, cert_bundle, cert_raw = self._certrequest_approve(request_id)
            else:
                poll_identifier = request_id
        elif request_status == "complete":
            error, cert_bundle, cert_raw = self._handle_enroll_complete(request_id)
        elif request_status == "rejected":
            self.logger.error("Certificate request was rejected by CA")
            error = "Certificate request was rejected by CA."
        else:
            self.logger.error(
                "Certificate request failed. Unknown request-status: %s",
                request_status,
            )
            error = "Certificate request failed"

        return (error, cert_bundle, cert_raw, poll_identifier)

    def _handle_enroll_complete(
        self, request_id: Optional[str]
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Handle the 'complete' status for enrollment.

        Args:
            request_id (Optional[str]): The request ID returned from the CA.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str]]: Error, PEM bundle, raw certificate.
        """
        code, response = self._api_get(f"{self.REST_CERTREQUESTS}/{request_id}")
        if code == 200 and isinstance(response, dict) and "certId" in response:
            return self._fetch_cert_and_bundle(response["certId"])
        else:
            self.logger.error(
                "Failed to retrieve certificate data for completed request. Status code: %s, Response: %s",
                code,
                response,
            )
            error = "Failed to retrieve certificate data for completed request."
            return error, None, None

    def handler_check(self) -> Optional[str]:
        """
        Check if handler is ready and configuration is valid.

        Returns:
            Optional[str]: Error message if any, else None.
        """
        self.logger.debug("CAhandler.check()")
        error = handler_config_check(self.logger, self, ["api_host", "client_cert"])
        self.logger.debug("CAhandler.check() ended with %s", error)
        return error

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str], bool]:
        """
        Poll status of pending CSR and download certificates.

        Args:
            _cert_name (str): Certificate name.
            poll_identifier (str): Poll identifier.
            _csr (str): CSR string (unused).
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str], Optional[str], bool]: Error, PEM bundle, raw certificate, poll identifier, rejected flag.
        """
        self.logger.debug("CAhandler.poll()")
        error: Optional[str] = None
        cert_bundle: Optional[str] = None
        cert_raw: Optional[str] = None
        rejected: bool = False
        # This method is a stub and should be implemented as needed.
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self,
        cert: str,
        _rev_reason: Optional[str] = None,
        _rev_date: Optional[str] = None,
    ) -> Tuple[int, str, str]:
        """
        Revoke a certificate.

        Args:
            cert (str): Certificate string.
            _rev_reason (Optional[str]): Revocation reason.
            _rev_date (Optional[str]): Revocation date.
        Returns:
            Tuple[int, str, str]: Code, message, and detail.
        """
        self.logger.debug("CAhandler.revoke()")

        if cert:
            serial = cert_serial_get(self.logger, cert, hexformat=True)

            if serial:
                code, message, detail = self._revoke(serial)
            else:
                self.logger.error(
                    "Failed to extract serial number from certificate for revocation"
                )
                code = 400
                message = "urn:ietf:params:acme:error:malformed"
                detail = "Invalid certificate format or missing serial number"
        else:
            self.logger.error("Certificate data is required for revocation")
            code = 400
            message = "urn:ietf:params:acme:error:malformed"
            detail = "Certificate data is required for revocation"

        self.logger.debug("Certificate.revoke() ended")
        return (code, message, detail)

    def trigger(
        self, _payload: str
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Process trigger message and return certificate.

        Args:
            _payload (str): Trigger payload.
        Returns:
            Tuple[Optional[str], Optional[str], Optional[str]]: Error, PEM bundle, and raw certificate.
        """
        self.logger.debug("CAhandler.trigger()")

        error: Optional[str] = None
        cert_bundle: Optional[str] = None
        cert_raw: Optional[str] = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""Test MS-WCCE CA handler connectivity."""

from __future__ import annotations

from acme2certifier.acme_srv.helper import logger_setup
from acme2certifier.cahandlers.mswcce_ca_handler import CAhandler


def main() -> None:
    """Create a test MS-WCCE request against the configured CA."""
    logger = logger_setup(True)
    with CAhandler(True, logger) as ca_handler:
        ca_handler.request_create()


if __name__ == "__main__":
    main()
